//*****************************************************************************
// File: ejava.cpp
// Description: interface to java
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifdef WIN32
#pragma warning( disable : 4996 )
#endif

#include "task.h"
#include "util.h"
#include "service.h"
#include "nvsint.h"
#include "nvcmd.h"
#include "ejava.h"
#include "global.h"
#include "../nvs/nvjava.h"
#include <map>
#include <vector>
#include <string>

using namespace std;

// Please uncomment the next line for debugging the solaris jvm problem under window
//#define HOTSPOT_DEBUG

// On SOLARIS, the libjava.so is not loaded with dlopen because SUN has reported that it doesn't work
// properly.
// So the library is implicitly linked

#ifdef SOLARIS
//#define JVM_IMPLICIT
#endif
#ifdef LINUX
//#define JVM_IMPLICIT
#endif

#ifdef JVM_IMPLICIT
#define CreateJavaVM JNI_CreateJavaVM
#else
typedef jint (JNICALL CreateJavaVM_t)(JavaVM **pvm, void **env, void *args);
#endif


// Global java virtual machine
JavaVM *jvm = NULL;

// And pointer to the JNI_CreateJavaVM function in the jvm library
#ifndef JVM_IMPLICIT
CreateJavaVM_t *CreateJavaVM = NULL;
#endif

// And finally the handle to the jvm library
#ifndef JVM_IMPLICIT
HMODULE hJvm = NULL;
#endif


// Unique global thread list object
NvJavaThreadList ThreadList;

// Unique global environment list object (used for java services)
NvJavaEnvList EnvList;

// Protection for Attach/Detach
NvMutex ProtectAttach;

// Nirva directory
string NirvaDirHere;

#ifdef WIN32
// Main dll entry point for windows
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
		
	return 1;   // ok
}
#endif

bool MyAttachCurrentThread(void **JavaEnv)
{
	if(jvm == NULL)
		return false;
	if(JavaEnv == NULL)
		return false;
	bool RetVal = true;
//	ProtectAttach.Lock();
	try
	{
//		printf("AAAAAttach Thread %d\n", GetCurrentThreadId());
		jint res = jvm->AttachCurrentThread(JavaEnv, NULL);
		if(res != 0)
		{
			printf("*****\n*****\n*****\n*****\n*****\nError in AttachCurrentThread\n*****\n*****\n*****\n*****\n*****");
			RetVal = false;
		}
	}
	catch(...)
	{
		printf("*****\n*****\n*****\n*****\n*****\nException in AttachCurrentThread\n*****\n*****\n*****\n*****\n*****");
		RetVal = false;
	}
	if(JavaEnv == NULL)
	{
		printf("*****\n*****\n*****\n*****\n*****\nNull environment returned by AttachCurrentThread\n*****\n*****\n*****\n*****\n*****");
		RetVal = false;
	}
//	ProtectAttach.Unlock();
	return RetVal;
}


void MyDetachCurrentThread()
{
	if(jvm == NULL)
		return;
//	ProtectAttach.Lock();
	try
	{
		jvm->DetachCurrentThread();
//		printf("DDDDDetach Thread %d\n", GetCurrentThreadId());
	}
	catch(...)
	{
		printf("*****\n*****\n*****\n*****\n*****\nException in DetachCurrentThread\n*****\n*****\n*****\n*****\n*****");
	}
//	ProtectAttach.Unlock();
}


NvJavaThreadList::NvJavaThreadList()
{
	// Do nothing in the constructor
}


NvJavaThreadList::~NvJavaThreadList()
{
	RemoveAll();
}


// Remove all threads from the list
void NvJavaThreadList::RemoveAll()
{
	Lock();
	map<NVTHREADID, NvJavaInt*>::iterator pos;
	for(pos = m_mList.begin(); pos != m_mList.end(); ++pos)
	{
		NvJavaInt* JavaInt = pos->second;
		if(JavaInt != NULL)
			delete JavaInt;
	}
	m_mList.clear();
	Unlock();
}


// Eventually attach the current thread and retuns a NvJavaInt class that owns the thread context
// This function is called from the RunMethod function
// If the pCommand parameter is NULL, the function just try to get the thread context
NvJavaInt* NvJavaThreadList::GetThreadNvIntPtr(void* pCommand)
{
	NvJavaInt* RetVal = NULL;
	NVTHREADID ThreadHandle;
#ifdef WIN32
	ThreadHandle = GetCurrentThreadId();
#else
	ThreadHandle = pthread_self();
#endif

	// This function must be thread protected
	Lock();
	
	// Try to see if the current thread has ever been attached
	// At this time, it exists in the list
	if(m_mList.find(ThreadHandle) != m_mList.end())
		RetVal = m_mList[ThreadHandle];

	if(pCommand == NULL)
	{
		// This is just a get, there is no creation
		Unlock();
		return RetVal;
	}

	if(RetVal == NULL)
	{
		// This is the first thread attachement
		// We must create a new thread context
		RetVal = new NvJavaInt;
		if(RetVal == NULL)
		{
			Unlock();
			return NULL;	// Memory error
		}

		// And attach the thread
		JNIEnv *env;
		if(!MyAttachCurrentThread((void **)(&env)))
		{
			Unlock();
			delete RetVal;
			return NULL;
		}
		
		// We must add the current thread to the list
		m_mList[ThreadHandle] = RetVal;

		// For performance reason we increase the in use counter
		// so the java thread is attached from the first use to the end of the thread
		// THis has been removed because produces increase of memory use and handles
		//RetVal->InUse++;

	}

	// We increment the the in use counter
	RetVal->InUse++;

	// We push the given command to the stack
	RetVal->m_mCommands.push_back(pCommand);

	// And set the actual command to the given one
	RetVal->m_pCommand = pCommand;

	Unlock();
	
	return RetVal;
}


// Eventually detach the current thread and remove the thread context
// If force is set to yes, we effectively detach in any case
void NvJavaThreadList::FreeThreadNvIntPtr(bool force)
{
	NVTHREADID ThreadHandle;
#ifdef WIN32
	ThreadHandle = GetCurrentThreadId();
#else
	ThreadHandle = pthread_self();
#endif
	
	// This function must be thread protected
	Lock();
	
	// If the thread is not in the list of attached thread
	// We do nothing except if force required
	if(m_mList.find(ThreadHandle) == m_mList.end())
	{
		if(force)
			MyDetachCurrentThread();
		Unlock();
		return;
	}

	// Get the thread context
	NvJavaInt* JavaInt = m_mList[ThreadHandle];
	if(JavaInt == NULL)
	{
		// This should not occur
		m_mList.erase(ThreadHandle);
		Unlock();
		return;
	}

	// We must eventually decrease the in use counter
	if(JavaInt->InUse > 0)
	{
		JavaInt->InUse--;
		if(JavaInt->m_mCommands.size())
		{
			// Should be always the case

			// We pop the actual command from the command stack
			JavaInt->m_mCommands.pop_back();

			// The actual command becomes the previous one if there is one
			if(JavaInt->m_mCommands.size())
				JavaInt->m_pCommand = JavaInt->m_mCommands[JavaInt->m_mCommands.size()-1];
			else
				JavaInt->m_pCommand = NULL;
		}
	}

	// if force mode or if this is the last use, we must detach the thread
	if(force || (JavaInt->InUse == 0))
	{
		MyDetachCurrentThread();
		// We also delete the thread context
		delete JavaInt;

		// And remove the thread from the list
		m_mList.erase(ThreadHandle);
	}

	Unlock();
}





NvJavaInt::NvJavaInt()
{
	ResultBufferSize = 0;
	SetBufferSize(0);
	LastResult = "";
	InUse = 0;
	m_pCommand = NULL;
}


NvJavaInt::~NvJavaInt()
{
	// Do nothing in the destructor
	if(ResultBufferSize)
		delete ResultBuffer;
}

int NvJavaInt::SetBufferSize(int DataSize)
{
	if(DataSize < ResultBufferSize)
		return 1;

	if(ResultBufferSize == 0)
	{
		ResultBuffer = (char *)malloc(RESULT_BUFFER_LENGTH);
		if(ResultBuffer == NULL)
			return 0;
		ResultBufferSize = RESULT_BUFFER_LENGTH;
	}

	if(DataSize < ResultBufferSize)
		return 1;

	char *NewResultBuffer = (char *)realloc(ResultBuffer, DataSize+1);
	if(NewResultBuffer == NULL)
		return 0;

	ResultBuffer = NewResultBuffer;
	ResultBufferSize = DataSize+1;

	return 2;
}






NvJavaEnv::NvJavaEnv()
{
	MyLoader = NULL;
	ClassToRun = NULL;
	ms = NULL;
	CurrentMethod = "";
	bStaticMethod = true;
	bOkOne = true;
	msLoaderFindClass = NULL;
	msRender = NULL;
}


NvJavaEnv::~NvJavaEnv()
{
	if(jvm != NULL)
	{
		// Attention, the env member must be valid before calling the destructor
		UnLoadClass();
	}
}


int NvJavaEnv::LoadClass(string DirName, string ClassName)
{

	// First unload the class if ever loaded
	UnLoadClass();

	JNIEnv *env;
	MyAttachCurrentThread((void **)(&env));
	if(env == NULL)
		return 0;

	vector<string> JarList;
	string JarDir = DirName;
	string::size_type index;
	index = JarDir.rfind('\\');
	if(index == string::npos)
		index = JarDir.rfind('/');
	if(index == string::npos)
		JarDir = "";
	else
		JarDir = JarDir.substr(0, index);
	if(JarDir.length())
	{
		vector<string> List;
		FileList(JarDir.c_str(), NULL, &List);
		int i;
		for(i=0; i<(int)List.size(); i++)
		{
			if(List[i].rfind(".jar") != string::npos) 
				JarList.push_back(List[i]);
		}
	}
	AddDirSuffix(JarDir);
	
	string URLName = "file:";
	URLName += DirName;
	jstring sURLName = env->NewStringUTF(URLName.c_str());
	jstring sClassName = env->NewStringUTF(ClassName.c_str());

	jclass nvjURLClass = env->FindClass("java/net/URL");
	if(nvjURLClass == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		return 0;
	}

	// We get its constructor method
	jmethodID msURLConstructor = env->GetMethodID(nvjURLClass, "<init>","(Ljava/lang/String;)V");
	if(msURLConstructor == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		return 0;
	}

	jobject MyURL = env->NewObject(nvjURLClass, msURLConstructor, sURLName);
	if(MyURL == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		return 0;
	}

	jclass nvjloaderClass = env->FindClass("java/net/URLClassLoader");
	if(nvjloaderClass == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		return 0;
	}

	// We get its constructor method
	jmethodID msLoaderConstructor = env->GetMethodID(nvjloaderClass, "<init>","([Ljava/net/URL;)V");
	if(msLoaderConstructor == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		return 0;
	}


	int UrlArraySize = 1;
	UrlArraySize += JarList.size();
	jobjectArray URLArray = env->NewObjectArray(UrlArraySize, nvjURLClass, MyURL);	
	int i;
	for(i=0; i<(int)JarList.size(); i++)
	{
		string JarFile = "file:";
		JarFile += JarDir;
		JarFile += JarList[i];
		jobject URLInList = env->NewObject(nvjURLClass, msURLConstructor, env->NewStringUTF(JarFile.c_str()));
		env->SetObjectArrayElement(URLArray, i+1, URLInList);
	}

	// Add the nirva.jar to the path of the new loader
/*	{
		string JarFile = "file:";
		JarFile += NirvaDirHere;
		JarFile += "Bin/nirva.jar";
		jobject URLInList = env->NewObject(nvjURLClass, msURLConstructor, env->NewStringUTF(JarFile.c_str()));
		env->SetObjectArrayElement(URLArray, i+1, URLInList);
	}
*/	
	// We create an instance of it
	jobject LocalObject = env->NewObject(nvjloaderClass, msLoaderConstructor, URLArray);
	if(LocalObject == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		return 0;
	}
	MyLoader = env->NewGlobalRef(LocalObject);
	env->DeleteLocalRef(LocalObject);

	
	// Get the findClass method 
	msLoaderFindClass = env->GetMethodID(nvjloaderClass, "findClass","(Ljava/lang/String;)Ljava/lang/Class;");
	if(msLoaderFindClass == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		env->DeleteGlobalRef(MyLoader);
		MyLoader = NULL;
		return 0;
	}

	jclass ClassObject = (jclass)(env->CallObjectMethod(MyLoader, msLoaderFindClass, sClassName));
	if(ClassObject == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		env->DeleteGlobalRef(MyLoader);
		MyLoader = NULL;
		return 0;
	}

	jmethodID msClassConstructor = env->GetMethodID(ClassObject, "<init>","()V");
	if(msClassConstructor == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		env->DeleteGlobalRef(MyLoader);
		MyLoader = NULL;
		return 0;
	}
	
	LocalObject = env->NewObject(ClassObject, msClassConstructor);
	if(LocalObject == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		env->DeleteGlobalRef(MyLoader);
		MyLoader = NULL;
		return 0;
	}
	ClassToRun = env->NewGlobalRef(LocalObject);
	env->DeleteLocalRef(LocalObject);

	return 1;
}


void NvJavaEnv::UnLoadClass()
{
	JNIEnv *env;
	MyAttachCurrentThread((void **)(&env));
	if(env == NULL)
		return;

	if(ClassToRun != NULL)
		env->DeleteGlobalRef(ClassToRun);
	if(MyLoader != NULL)
		env->DeleteGlobalRef(MyLoader);

	MyLoader = NULL;
	ClassToRun = NULL;
	ms = NULL;
	msRender = NULL;
}


int NvJavaEnv::RunMethod(std::string Method, bool BooleanMethod)
{
	JNIEnv *env;
	MyAttachCurrentThread((void **)(&env));
	if(env == NULL)
		return 0;

	if((ms == NULL) || (Method != CurrentMethod))
	{
		bStaticMethod = false;
		jclass cls = env->GetObjectClass(ClassToRun);
		if(BooleanMethod)
			ms = env->GetMethodID(cls, Method.c_str(),"()Z"); 
		else
			ms = env->GetMethodID(cls, Method.c_str(),"()I"); 
		if(ms == NULL)	// Maybe it's a static method
		{
			if (env->ExceptionOccurred())
				env->ExceptionClear();
			if(BooleanMethod)
				ms = env->GetStaticMethodID(cls, Method.c_str(),"()Z"); 
			else
				ms = env->GetStaticMethodID(cls, Method.c_str(),"()I"); 
			if(ms != NULL)
				bStaticMethod = true;
			else
			{
				if (env->ExceptionOccurred())
					env->ExceptionClear();
			}
		}
		if(ms == NULL)
		{
			// We cannot find the method
			// We must try with Main or main if nothing or Main or main was requested
			bool RetryIt = false;
			if(Method == "main")
			{
				// main was requested so we try with Main
				Method = "Main";
				RetryIt = true;
			}
			else if(Method == "Main")
			{
				// Main was requested so we try with main
				Method = "main";
				RetryIt = true;
			}
			if(RetryIt)
			{
				if(BooleanMethod)
					ms = env->GetMethodID(cls, Method.c_str(),"()Z"); 
				else
					ms = env->GetMethodID(cls, Method.c_str(),"()I"); 
				if(ms == NULL)	// Maybe it's a static method
				{
					if (env->ExceptionOccurred())
						env->ExceptionClear();
					if(BooleanMethod)
						ms = env->GetStaticMethodID(cls, Method.c_str(),"()Z"); 
					else
						ms = env->GetStaticMethodID(cls, Method.c_str(),"()I"); 
					if(ms != NULL)
						bStaticMethod = true;
					else
					{
						if (env->ExceptionOccurred())
							env->ExceptionClear();
					}
				}
			}
			if(ms == NULL)
				return 0;
		}
		CurrentMethod = Method;
	}

	// We got the method so we can run it
	// Run the command here
	int RetVal = 1;
	int RetCode = 0;
	try
	{
		if(bStaticMethod)
		{
			jclass cls = env->GetObjectClass(ClassToRun);
			if(BooleanMethod)
				RetCode = env->CallStaticBooleanMethod(cls, ms);
			else
				RetCode = env->CallStaticIntMethod(cls, ms);
		}
		else
		{
			if(BooleanMethod)
				RetCode = env->CallBooleanMethod(ClassToRun, ms);
			else
				RetCode = env->CallIntMethod(ClassToRun, ms);
		}
	}
	catch(...)
	{
		RetVal = 0;
	}

// Next line is commented because it may crash		
//	env->ExceptionClear();		

	if(bOkOne)
	{
		if(RetCode != 1)
			RetVal = 0;
	}
	else
	{
		if(RetCode != 0)
			RetVal = 0;
	}

	return RetVal;
}



bool NvJavaEnv::DoRender(NvRendererInt *Render)
{
	Render->SetReason("");

	JNIEnv *env;
	MyAttachCurrentThread((void **)(&env));
	if(env == NULL)
	{
		Render->SetReason("Cannot get a valid java environment");
		return false;
	}

	if(msRender == NULL)
	{
		jclass cls = env->GetObjectClass(ClassToRun);
		msRender = env->GetMethodID(cls, "Render","(Lcom/nirvasoft/nirva/nvrend;)Z"); 
	}

	if(msRender == NULL)
	{
		Render->SetReason("This is not a renderer service");
		return false;
	}

	if(msLoaderFindClass == NULL)
	{
		Render->SetReason("No find class method defined for the loader");
		return false;
	}

//	string ClassName = "com/nirvasoft/nirva/nvrend";
//	jstring sClassName = env->NewStringUTF(ClassName.c_str());

//	jclass ClassObject = (jclass)(env->CallObjectMethod(MyLoader, msLoaderFindClass, sClassName));
	jclass ClassObject = env->FindClass("com/nirvasoft/nirva/nvrend");
	if(ClassObject == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		Render->SetReason("Cannot find the nvrend class");
		return false;
	}

	jmethodID msClassConstructor = env->GetMethodID(ClassObject, "<init>","([BLjava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V");
	if(msClassConstructor == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		Render->SetReason("Cannot get nvrend class constructor");
		return false;
	}
	
	// We must now prepare parameters for constructor
	// public nvrend(byte[] InputBytes, java.lang.String sPageName, java.lang.String sSessionId)
	jstring sPageName = env->NewStringUTF(Render->GetPageName());
	jstring sSessionId = env->NewStringUTF(Render->GetSessionId());
	jstring sAppName = env->NewStringUTF(Render->GetAppName());
	jstring sInputEncoding = env->NewStringUTF(Render->GetInputEncoding());
	jbyteArray InputBytes = env->NewByteArray(Render->GetInputDataSize());
	env->SetByteArrayRegion(InputBytes, 0, Render->GetInputDataSize(), (const jbyte*)(Render->GetInputData()));

	// The input data has been transfered to Java so we can remove it from C++ in order to save memory
	Render->FreeInputData();

	jobject NvRendObject = env->NewObject(ClassObject, msClassConstructor, InputBytes, sPageName, sSessionId, sAppName, sInputEncoding);
	if(NvRendObject == NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
		env->DeleteLocalRef(InputBytes);
		Render->SetReason("Cannot create nvrend class instance");
		return false;
	}

	// Execute the Render method
	bool RetVal = true;
	bool ExceptionOccurred = false;
	try
	{
		if(!env->CallBooleanMethod(ClassToRun, msRender, NvRendObject))
			RetVal = false;
	}
	catch(...)
	{
		Render->SetReason("Exception occured in Render method");
		RetVal = false;
		ExceptionOccurred = true;
	}

	if(ExceptionOccurred)
	{
		env->DeleteLocalRef(InputBytes);
		env->DeleteLocalRef(NvRendObject);
		return false;
	}

	// Get the reason
	if(!RetVal)
	{
		jfieldID ReasonId = env->GetFieldID(ClassObject, "Reason", "Ljava/lang/String;");
		if(ReasonId != NULL)
		{
			jstring jReason = (jstring)(env->GetObjectField(NvRendObject, ReasonId));
			const char *str = env->GetStringUTFChars(jReason, 0);
			string ReasonHere = str;
			env->ReleaseStringUTFChars(jReason, str);
			Render->SetReason(ReasonHere.c_str());
		}
	}

	// Get the content-type
	if(RetVal)
	{
		jfieldID ContentTypeId = env->GetFieldID(ClassObject, "ContentType", "Ljava/lang/String;");
		if(ContentTypeId != NULL)
		{
			jstring jContentType = (jstring)(env->GetObjectField(NvRendObject, ContentTypeId));
			const char *str = env->GetStringUTFChars(jContentType, 0);
			string ContentTypeHere = str;
			env->ReleaseStringUTFChars(jContentType, str);
			Render->SetContentType(ContentTypeHere.c_str());
		}
	}

	// Get the output stream
	if(RetVal)
	{
		jmethodID msGetOutputBuffer = env->GetMethodID(ClassObject, "GetOutputBuffer","()[B");
		if(msGetOutputBuffer == NULL)
		{
			Render->SetReason("Cannot get java render output buffer");
			env->DeleteLocalRef(InputBytes);
			env->DeleteLocalRef(NvRendObject);
			return false;
		}

		jbyteArray OutputBytes = (jbyteArray)(env->CallObjectMethod(NvRendObject, msGetOutputBuffer));
		if(OutputBytes != NULL)
		{
			unsigned long Length = env->GetArrayLength(OutputBytes);
			char *RenderDataBytes = Render->AllocOutputBuffer(Length);
			env->GetByteArrayRegion(OutputBytes, 0, Length, (jbyte *)RenderDataBytes);
			env->DeleteLocalRef(OutputBytes);
		}
	}

	// Cleanup
	env->DeleteLocalRef(InputBytes);
	env->DeleteLocalRef(NvRendObject);
	return RetVal;
}




NvJavaEnvList::NvJavaEnvList()
{
	// Do nothing in the constructor
}


NvJavaEnvList::~NvJavaEnvList()
{
}


// Remove all items from the list
void NvJavaEnvList::RemoveAll()
{
	Lock();
	map<std::string, NvJavaEnv*>::iterator pos;
	for(pos = m_mList.begin(); pos != m_mList.end(); ++pos)
	{
		NvJavaEnv* JavaEnv = pos->second;
		if(JavaEnv != NULL)
		{
			delete JavaEnv;
		}
	}
	m_mList.clear();
	Unlock();
}


// Get the NvJavaEnv corresponding to the service
NvJavaEnv* NvJavaEnvList::GetJavaEnv(string ServiceName)
{
	NvJavaEnv* RetVal = NULL;

	// This function must be thread protected
	Lock();
	
	if(m_mList.find(ServiceName) != m_mList.end())
		RetVal = m_mList[ServiceName];

	Unlock();
	return RetVal;
}


// Add the NvJavaEnv corresponding to the service
void NvJavaEnvList::AddJavaEnv(string ServiceName, NvJavaEnv* JavaEnv)
{
	// This function must be thread protected
	Lock();
	
	if(m_mList.find(ServiceName) != m_mList.end())
	{
		NvJavaEnv* JavaEnvInList = m_mList[ServiceName];
		if(JavaEnvInList != NULL)
		{
			delete JavaEnvInList;
		}
	}
	m_mList[ServiceName] = JavaEnv;
	Unlock();
}


// Remove the NvJavaEnv corresponding to the service
void NvJavaEnvList::RemoveJavaEnv(string ServiceName)
{
	// This function must be thread protected
	Lock();
	
	if(m_mList.find(ServiceName) != m_mList.end())
	{
		NvJavaEnv* JavaEnvInList = m_mList[ServiceName];
		if(JavaEnvInList != NULL)
		{
			delete JavaEnvInList;
		}
		m_mList.erase(ServiceName);
	}
	Unlock();
}





// Library exported functions

// Loads the JVM
// The load and unload of the JVM must be made from the same thread
int NvInitJava(const char* NirvaDir, char* MinMemSize, char* MaxMemSize, const char* JavaHomeDir, const char* JvmPath, const char* JvmParam)
{

	// Do nothing if ever loaded
	if(jvm != NULL)
		return 1;
	if(NirvaDir == NULL)
		return 0;
	
	NirvaDirHere = NirvaDir;
	AddDirSuffix(NirvaDirHere);
	string LibFile = NirvaDir;
	RemoveDirSuffix(NirvaDirHere);

// The next line is for sun jre
//	LibFile += "Java/bin/server/jvm.dll";
#ifdef WIN32
//	LibFile += "Java/bin/client/jvm.dll";
	LibFile += "Java/bin/server/jvm.dll";
#endif
#ifdef AIX
	LibFile += "Java/bin/libjava.a";
#endif
#ifdef LINUX
//	LibFile += "Java/bin/libjava.so";
#ifdef X64
	LibFile += "Java/lib/amd64/libjava.so";
#else
	LibFile += "Java/lib/i386/libjava.so";
#endif
#endif
#ifdef HPUX
	LibFile += "Java/lib/PA_RISC2.0/server/libjvm.sl";
#endif
#ifdef HPUXI
	LibFile += "Java/lib/IA64N/server/libjvm.so";
#endif
#ifdef SOLARIS
//	LibFile += "Java/lib/sparc/libjava.so";
	LibFile += "Java/lib/sparc/server/libjvm.so";
#endif
	if((JvmPath!=NULL) && strlen(JvmPath))
		LibFile=JvmPath;
	SetDirName(LibFile);


	string JavaHomeDirHere = NirvaDir;
	AddDirSuffix(NirvaDirHere);
	JavaHomeDirHere += "Java";
	if((JavaHomeDir!=NULL) && strlen(JavaHomeDir))
		JavaHomeDirHere = JavaHomeDir;
	string JavaLibDir = JavaHomeDirHere;
	AddDirSuffix(JavaLibDir);
	JavaLibDir += "lib";
	SetDirName(JavaHomeDirHere);
	SetDirName(JavaLibDir);

#ifndef JVM_IMPLICIT

	//Load the Java VM DLL
	hJvm = MyLoadLibrary(LibFile.c_str());
	if(hJvm == NULL)
		return 0;

	// Get the JNI_CreateJavaVM function pointer
	CreateJavaVM = (CreateJavaVM_t *)MyGetLibFunction(hJvm, "JNI_CreateJavaVM");
	if (CreateJavaVM == NULL)
	{
		MyFreeLibrary(hJvm);
		hJvm = NULL;
		return 0;
	}
#endif
	// Prepare the options for the JVM
	JNIEnv *env;
	JavaVMInitArgs vm_args;

	JavaVMOption options[110];
	vector<char *> AdditionnalParams;
	char *JvmParamBuffer = NULL;
	if((JvmParam != NULL) && strlen(JvmParam))
	{
		// Parse jvm extra parameters
		JvmParamBuffer = (char *)malloc(strlen(JvmParam)+10);
		strcpy(JvmParamBuffer, JvmParam);
		char *Parameter = (char *)JvmParamBuffer;
		char *Value;
		int pos;
		// Get first real character
		while(strlen(Parameter))
		{
			Value = Parameter;
			while(Value[0]==' ')
				Value++;
			pos = 0;
			if(Value[0] == '\"')
			{
				Value++;
				while(1)
				{
					if((Value[pos] == '\"') && (Value[pos-1] != '\\'))
					{
						Value[pos] = '\0';
						pos++;
						break;
					}
					pos++;
				}
			}
			else
			{
				while(1)
				{
					if(Value[pos] == ' ')
					{
						Value[pos] = '\0';
						pos++;
						break;
					}
					if(Value[pos] == '\0')
						break;
					pos++;
				}
			}
			Parameter = Value+pos;
			if(strlen(Value))
			{
				string ValueHere = Value;
				// Remove reserved Jvm options
				MakeUpper(ValueHere);
				if(ValueHere.find("-DJAVA.CLASS.PATH=") == 0)
					continue;
				if(ValueHere.find("-DJAVA.HOME=") == 0)
					continue;
				if(ValueHere == "-XRS")
					continue;
				if(ValueHere.find("-XMS") == 0)
					continue;
				if(ValueHere.find("-XMX") == 0)
					continue;
				ValueHere = Value;
				AdditionnalParams.push_back(Value);
				if(AdditionnalParams.size() >= 100)
					break;		// limited to 100 options
			}
		}
	}
	
	// Set the main classpath
	string sClassPath = "-Djava.class.path=";
	// Nirva dir
	sClassPath += NirvaDirHere;
	// Nirva jar first
	sClassPath += PATH_SEPARATOR;
	sClassPath += NirvaDirHere;
	sClassPath += "Bin/nirva.jar";
	sClassPath += PATH_SEPARATOR;
	sClassPath += JavaLibDir;
	// Nirva/Bin dir
	sClassPath += PATH_SEPARATOR;
	sClassPath += NirvaDirHere;
	sClassPath += "Bin";

	string JarDir = NirvaDirHere;
	JarDir += "Bin";
	vector<string> List;
	FileList(JarDir.c_str(), NULL, &List);
	AddDirSuffix(JarDir);
	int j;
	for(j=0; j<(int)List.size(); j++)
	{
		if(List[j].rfind(".jar") != string::npos)
		{
			sClassPath += PATH_SEPARATOR;
			sClassPath += JarDir;
			sClassPath += List[j];
		}
	}


	// General CLASSPATH
	char *OfficialClassPath = getenv("CLASSPATH");
	if(OfficialClassPath != NULL)
	{
		sClassPath += PATH_SEPARATOR;
		sClassPath += OfficialClassPath;
	}
	SetDirName(sClassPath);
	char *ClassPathBuffer = (char *)malloc(sClassPath.size()+10);
	if(ClassPathBuffer == NULL)
	{
		if(JvmParamBuffer != NULL)
			free(JvmParamBuffer);
		return 0;
	}

	strcpy(ClassPathBuffer, sClassPath.c_str());
	options[0].optionString = ClassPathBuffer;
	int NumOptions = 1;
	
	
	// Set the java home directory
	string sJavaPath = "-Djava.home=";
	sJavaPath += JavaHomeDirHere;
	SetDirName(sJavaPath);
	char *JavaPathBuffer = (char *)malloc(sJavaPath.size()+10);
	if(JavaPathBuffer == NULL)
	{
		if(JvmParamBuffer != NULL)
			free(JvmParamBuffer);
		free(ClassPathBuffer);
		return 0;
	}
	strcpy(JavaPathBuffer, sJavaPath.c_str());
	options[NumOptions].optionString = JavaPathBuffer;
	NumOptions++;


#ifndef HOTSPOT_DEBUG
	// The Wrs option allows to remove many signal management from the JVM
	options[NumOptions].optionString = (char *)"-Xrs";
	NumOptions++;
#endif

	// Set heap size options
	if(MinMemSize != NULL)
	{
		options[NumOptions].optionString = MinMemSize;
		NumOptions++;
	}
	if(MaxMemSize != NULL)
	{
		options[NumOptions].optionString = MaxMemSize;
		NumOptions++;
	}

	// Add extra options
	for(j=0; j<(int)AdditionnalParams.size(); j++)
	{
		options[NumOptions].optionString = AdditionnalParams[j];
		NumOptions++;
	}

//#ifdef NVDEBUG
//	options[NumOptions].optionString = "-verbose:jni";
//	NumOptions++;
//	options[NumOptions].optionString = "-verbose:class";
//	NumOptions++;
//	options[NumOptions].optionString = "-verbose:gc";
//	NumOptions++;

//	options[NumOptions].optionString = "-Xss128k";
//	NumOptions++;
	
//	options[NumOptions].optionString = "-Xdebug";
//	NumOptions++;
	
//	options[NumOptions].optionString = "-Xrunjdwp:transport=dt_shmem,address=nirva,server=y,suspend=n";
//	NumOptions++;

//	options[NumOptions].optionString = "-Xnoclassgc";
//	NumOptions++;
	
// Uncomment next line for using profiler (windows)
//	options[NumOptions].optionString = "-Xrunjmp";
//	NumOptions++;

//#endif


	// We use the 1.4 version
	// Except on AIX if using Nirva jvm
	vm_args.version = JNI_VERSION_1_4;
#ifdef AIX
	if((JvmPath==NULL) || (!strlen(JvmPath)))
		vm_args.version = JNI_VERSION_1_2;
#endif
	vm_args.options = options;
	vm_args.nOptions = NumOptions;
	vm_args.ignoreUnrecognized = JNI_FALSE;

	// Really create the JVM here
	jint res = CreateJavaVM(&jvm, (void **)&env, &vm_args);
	if(res < 0)
	{
		if(JvmParamBuffer != NULL)
			free(JvmParamBuffer);
		free(ClassPathBuffer);
		free(JavaPathBuffer);
#ifndef JVM_IMPLICIT
		MyFreeLibrary(hJvm);
		hJvm = NULL;
#endif
		jvm = NULL;
		return 0;
	}

	// We don't need that any more
	if(JvmParamBuffer != NULL)
		free(JvmParamBuffer);
	free(ClassPathBuffer);
	free(JavaPathBuffer);


	// We must now register the NIRVA interface functions
	// This is a class named com.nirvasoft.nirva.nvsint in the nirva.jar file in Nirva/Bin directory)
	jclass nvcls = env->FindClass("com/nirvasoft/nirva/nvsint");
	if(nvcls == NULL)
	{
		jvm->DestroyJavaVM();
		jvm = NULL;
#ifndef JVM_IMPLICIT
		MyFreeLibrary(hJvm);
		hJvm = NULL;
#endif
		return 0;
	}


	// GetCommandHandle function
	JNINativeMethod *methods = new JNINativeMethod[18];
	methods[0].name = (char *)"Int_GetCommandHandle";
	methods[0].signature = (char *)"()J";
	methods[0].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetCommandHandle;

	// SetError function
	methods[1].name = (char *)"Int_SetError";
	methods[1].signature = (char *)"(JLjava/lang/String;)V";
	methods[1].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1SetError;

	// SetErrorEx function
	methods[2].name = (char *)"Int_SetErrorEx";
	methods[2].signature = (char *)"(JLjava/lang/String;Ljava/lang/String;ILjava/lang/String;)V";
	methods[2].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1SetErrorEx;

	// Command function
	methods[3].name = (char *)"Int_Command";
	methods[3].signature = (char *)"(JLjava/lang/String;)I";
	methods[3].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1Command;

	// GetResult function
	methods[4].name = (char *)"Int_GetResult";
	methods[4].signature = (char *)"(J)Ljava/lang/String;";
	methods[4].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetResult;

	// GetSessionId function
	methods[5].name = (char *)"Int_GetSessionId";
	methods[5].signature = (char *)"(J)Ljava/lang/String;";
	methods[5].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetSessionId;

	// GetService function
	methods[6].name = (char *)"Int_GetService";
	methods[6].signature = (char *)"(J)Ljava/lang/String;";
	methods[6].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetService;

	// GetClass function
	methods[7].name = (char *)"Int_GetClass";
	methods[7].signature = (char *)"(J)Ljava/lang/String;";
	methods[7].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetClass;

	// GetCommand function
	methods[8].name = (char *)"Int_GetCommand";
	methods[8].signature = (char *)"(J)Ljava/lang/String;";
	methods[8].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetCommand;

	// GetInContainer function
	methods[9].name = (char *)"Int_GetInContainer";
	methods[9].signature = (char *)"(J)Ljava/lang/String;";
	methods[9].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetInContainer;

	// GetOutContainer function
	methods[10].name = (char *)"Int_GetOutContainer";
	methods[10].signature = (char *)"(J)Ljava/lang/String;";
	methods[10].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetOutContainer;

	// GetLanguage function
	methods[11].name = (char *)"Int_GetLanguage";
	methods[11].signature = (char *)"(J)Ljava/lang/String;";
	methods[11].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetLanguage;

	// GetError function
	methods[12].name = (char *)"Int_GetError";
	methods[12].signature = (char *)"(JLjava/lang/String;)Ljava/lang/String;";
	methods[12].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetError;

	// GetNumParameters function
	methods[13].name = (char *)"Int_GetNumParameters";
	methods[13].signature = (char *)"(J)I";
	methods[13].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetNumParameters;

	// GetSource function
	methods[14].name = (char *)"Int_GetSource";
	methods[14].signature = (char *)"(J)Ljava/lang/String;";
	methods[14].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetSource;

	// ParameterExist function
	methods[15].name = (char *)"Int_ParameterExist";
	methods[15].signature = (char *)"(JLjava/lang/String;)I";
	methods[15].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1ParameterExist;

	// GetParameter function
	methods[16].name = (char *)"Int_GetParameter";
	methods[16].signature = (char *)"(JLjava/lang/String;)Ljava/lang/String;";
	methods[16].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetParameter;

	// GetParameterName function
	methods[17].name = (char *)"Int_GetParameterName";
	methods[17].signature = (char *)"(JI)Ljava/lang/String;";
	methods[17].fnPtr = (void*)Java_com_nirvasoft_nirva_nvsint_Int_1GetParameterName;

	// Do the registration here
	jint i = env->RegisterNatives(nvcls, methods, 18);

	// We don't need that any more
	delete methods;

	if(i != 0)
	{
		jvm->DestroyJavaVM();
		jvm = NULL;
#ifndef JVM_IMPLICIT
		MyFreeLibrary(hJvm);
		hJvm = NULL;
#endif
		return 0;
	}
/*
	// For compatibility, we also register unpackaged class 
	// This is a class named nvsint (file nvsint.class in Nirva/Bin directory)

	jclass nvcls2 = env->FindClass("nvsint");
	if(nvcls2 == NULL)
	{
		jvm->DestroyJavaVM();
		jvm = NULL;
#ifndef JVM_IMPLICIT
		MyFreeLibrary(hJvm);
		hJvm = NULL;
#endif
		return 0;
	}


	// GetCommandHandle function
	methods = new JNINativeMethod[18];
	methods[0].name = (char *)"Int_GetCommandHandle";
	methods[0].signature = (char *)"()I";
	methods[0].fnPtr = (void*)Java_nvsint_Int_1GetCommandHandle;

	// SetError function
	methods[1].name = (char *)"Int_SetError";
	methods[1].signature = (char *)"(ILjava/lang/String;)V";
	methods[1].fnPtr = (void*)Java_nvsint_Int_1SetError;

	// SetErrorEx function
	methods[2].name = (char *)"Int_SetErrorEx";
	methods[2].signature = (char *)"(ILjava/lang/String;Ljava/lang/String;ILjava/lang/String;)V";
	methods[2].fnPtr = (void*)Java_nvsint_Int_1SetErrorEx;

	// Command function
	methods[3].name = (char *)"Int_Command";
	methods[3].signature = (char *)"(ILjava/lang/String;)I";
	methods[3].fnPtr = (void*)Java_nvsint_Int_1Command;

	// GetResult function
	methods[4].name = (char *)"Int_GetResult";
	methods[4].signature = (char *)"(I)Ljava/lang/String;";
	methods[4].fnPtr = (void*)Java_nvsint_Int_1GetResult;

	// GetSessionId function
	methods[5].name = (char *)"Int_GetSessionId";
	methods[5].signature = (char *)"(I)Ljava/lang/String;";
	methods[5].fnPtr = (void*)Java_nvsint_Int_1GetSessionId;

	// GetService function
	methods[6].name = (char *)"Int_GetService";
	methods[6].signature = (char *)"(I)Ljava/lang/String;";
	methods[6].fnPtr = (void*)Java_nvsint_Int_1GetService;

	// GetClass function
	methods[7].name = (char *)"Int_GetClass";
	methods[7].signature = (char *)"(I)Ljava/lang/String;";
	methods[7].fnPtr = (void*)Java_nvsint_Int_1GetClass;

	// GetCommand function
	methods[8].name = (char *)"Int_GetCommand";
	methods[8].signature = (char *)"(I)Ljava/lang/String;";
	methods[8].fnPtr = (void*)Java_nvsint_Int_1GetCommand;

	// GetInContainer function
	methods[9].name = (char *)"Int_GetInContainer";
	methods[9].signature = (char *)"(I)Ljava/lang/String;";
	methods[9].fnPtr = (void*)Java_nvsint_Int_1GetInContainer;

	// GetOutContainer function
	methods[10].name = (char *)"Int_GetOutContainer";
	methods[10].signature = (char *)"(I)Ljava/lang/String;";
	methods[10].fnPtr = (void*)Java_nvsint_Int_1GetOutContainer;

	// GetLanguage function
	methods[11].name = (char *)"Int_GetLanguage";
	methods[11].signature = (char *)"(I)Ljava/lang/String;";
	methods[11].fnPtr = (void*)Java_nvsint_Int_1GetLanguage;

	// GetError function
	methods[12].name = (char *)"Int_GetError";
	methods[12].signature = (char *)"(ILjava/lang/String;)Ljava/lang/String;";
	methods[12].fnPtr = (void*)Java_nvsint_Int_1GetError;

	// GetNumParameters function
	methods[13].name = (char *)"Int_GetNumParameters";
	methods[13].signature = (char *)"(I)I";
	methods[13].fnPtr = (void*)Java_nvsint_Int_1GetNumParameters;

	// GetSource function
	methods[14].name = (char *)"Int_GetSource";
	methods[14].signature = (char *)"(I)Ljava/lang/String;";
	methods[14].fnPtr = (void*)Java_nvsint_Int_1GetSource;

	// ParameterExist function
	methods[15].name = (char *)"Int_ParameterExist";
	methods[15].signature = (char *)"(ILjava/lang/String;)I";
	methods[15].fnPtr = (void*)Java_nvsint_Int_1ParameterExist;

	// GetParameter function
	methods[16].name = (char *)"Int_GetParameter";
	methods[16].signature = (char *)"(ILjava/lang/String;)Ljava/lang/String;";
	methods[16].fnPtr = (void*)Java_nvsint_Int_1GetParameter;

	// GetParameterName function
	methods[17].name = (char *)"Int_GetParameterName";
	methods[17].signature = (char *)"(II)Ljava/lang/String;";
	methods[17].fnPtr = (void*)Java_nvsint_Int_1GetParameterName;

	// Do the registration here
	i = env->RegisterNatives(nvcls2, methods, 18);

	// We don't need that any more
	delete methods;

	if(i != 0)
	{
		jvm->DestroyJavaVM();
		jvm = NULL;
#ifndef JVM_IMPLICIT
		MyFreeLibrary(hJvm);
		hJvm = NULL;
#endif
		return 0;
	}
*/

	// We register now service interface for java service
	jclass nvcls3 = env->FindClass("com/nirvasoft/nirva/nvcmd");
	if(nvcls3 == NULL)
	{
		jvm->DestroyJavaVM();
		jvm = NULL;
#ifndef JVM_IMPLICIT
		MyFreeLibrary(hJvm);
		hJvm = NULL;
#endif
		return 0;
	}


	// GetCommandHandle function
	methods = new JNINativeMethod[20];
	methods[0].name = (char *)"Int_GetCommandHandle";
	methods[0].signature = (char *)"()J";
	methods[0].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetCommandHandle;

	// SetError function
	methods[1].name = (char *)"Int_SetError";
	methods[1].signature = (char *)"(JLjava/lang/String;ILjava/lang/String;)V";
	methods[1].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1SetError;

	// Command function
	methods[2].name = (char *)"Int_Command";
	methods[2].signature = (char *)"(JLjava/lang/String;)Z";
	methods[2].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1Command;

	// GetResult function
	methods[3].name = (char *)"Int_GetResult";
	methods[3].signature = (char *)"(J)Ljava/lang/String;";
	methods[3].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetResult;

	// GetSessionId function
	methods[4].name = (char *)"Int_GetSessionId";
	methods[4].signature = (char *)"(J)Ljava/lang/String;";
	methods[4].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetSessionId;

	// GetClass function
	methods[5].name = (char *)"Int_GetClass";
	methods[5].signature = (char *)"(J)Ljava/lang/String;";
	methods[5].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetClass;

	// GetCommand function
	methods[6].name = (char *)"Int_GetCommand";
	methods[6].signature = (char *)"(J)Ljava/lang/String;";
	methods[6].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetCommand;

	// GetInContainer function
	methods[7].name = (char *)"Int_GetInContainer";
	methods[7].signature = (char *)"(J)Ljava/lang/String;";
	methods[7].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetInContainer;

	// GetOutContainer function
	methods[8].name = (char *)"Int_GetOutContainer";
	methods[8].signature = (char *)"(J)Ljava/lang/String;";
	methods[8].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetOutContainer;

	// GetLanguage function
	methods[9].name = (char *)"Int_GetLanguage";
	methods[9].signature = (char *)"(J)Ljava/lang/String;";
	methods[9].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetLanguage;

	// GetMachineName function
	methods[10].name = (char *)"Int_GetMachineName";
	methods[10].signature = (char *)"(J)Ljava/lang/String;";
	methods[10].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetMachineName;

	// GetMachineUser function
	methods[11].name = (char *)"Int_GetMachineUser";
	methods[11].signature = (char *)"(J)Ljava/lang/String;";
	methods[11].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetMachineUser;

	// GetError function
	methods[12].name = (char *)"Int_GetError";
	methods[12].signature = (char *)"(JLjava/lang/String;)Ljava/lang/String;";
	methods[12].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetError;

	// GetNumParameters function
	methods[13].name = (char *)"Int_GetNumParameters";
	methods[13].signature = (char *)"(J)I";
	methods[13].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetNumParameters;

	// GetParameterName function
	methods[14].name = (char *)"Int_GetParameterName";
	methods[14].signature = (char *)"(JI)Ljava/lang/String;";
	methods[14].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetParameterName;

	// GetSource function
	methods[15].name = (char *)"Int_GetSource";
	methods[15].signature = (char *)"(J)Ljava/lang/String;";
	methods[15].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetSource;

	// IsCommand function
	methods[16].name = (char *)"Int_IsCommand";
	methods[16].signature = (char *)"(JLjava/lang/String;Ljava/lang/String;Ljava/lang/String;)Z";
	methods[16].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1IsCommand;

	// ParameterExist function
	methods[17].name = (char *)"Int_ParameterExist";
	methods[17].signature = (char *)"(JLjava/lang/String;)Z";
	methods[17].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1ParameterExist;

	// GetParameter function
	methods[18].name = (char *)"Int_GetParameter";
	methods[18].signature = (char *)"(JLjava/lang/String;)Ljava/lang/String;";
	methods[18].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1GetParameter;

	// AddHttpHeader function
	methods[19].name = (char *)"Int_AddHttpHeader";
	methods[19].signature = (char *)"(JLjava/lang/String;Ljava/lang/String;)V";
	methods[19].fnPtr = (void*)Java_com_nirvasoft_nirva_nvcmd_Int_1AddHttpHeader;

	// Do the registration here
	i = env->RegisterNatives(nvcls3, methods, 20);

	// We don't need that any more
	delete methods;

	if(i != 0)
	{
		jvm->DestroyJavaVM();
		jvm = NULL;
#ifndef JVM_IMPLICIT
		MyFreeLibrary(hJvm);
		hJvm = NULL;
#endif
		return 0;
	}

/*
	// Same for old classes
	jclass nvcls4 = env->FindClass("nvcmd");
	if(nvcls4 == NULL)
	{
		jvm->DestroyJavaVM();
		jvm = NULL;
#ifndef JVM_IMPLICIT
		MyFreeLibrary(hJvm);
		hJvm = NULL;
#endif
		return 0;
	}


	// GetCommandHandle function
	methods = new JNINativeMethod[20];
	methods[0].name = (char *)"Int_GetCommandHandle";
	methods[0].signature = (char *)"()I";
	methods[0].fnPtr = (void*)Java_nvcmd_Int_1GetCommandHandle;

	// SetError function
	methods[1].name = (char *)"Int_SetError";
	methods[1].signature = (char *)"(ILjava/lang/String;ILjava/lang/String;)V";
	methods[1].fnPtr = (void*)Java_nvcmd_Int_1SetError;

	// Command function
	methods[2].name = (char *)"Int_Command";
	methods[2].signature = (char *)"(ILjava/lang/String;)Z";
	methods[2].fnPtr = (void*)Java_nvcmd_Int_1Command;

	// GetResult function
	methods[3].name = (char *)"Int_GetResult";
	methods[3].signature = (char *)"(I)Ljava/lang/String;";
	methods[3].fnPtr = (void*)Java_nvcmd_Int_1GetResult;

	// GetSessionId function
	methods[4].name = (char *)"Int_GetSessionId";
	methods[4].signature = (char *)"(I)Ljava/lang/String;";
	methods[4].fnPtr = (void*)Java_nvcmd_Int_1GetSessionId;

	// GetClass function
	methods[5].name = (char *)"Int_GetClass";
	methods[5].signature = (char *)"(I)Ljava/lang/String;";
	methods[5].fnPtr = (void*)Java_nvcmd_Int_1GetClass;

	// GetCommand function
	methods[6].name = (char *)"Int_GetCommand";
	methods[6].signature = (char *)"(I)Ljava/lang/String;";
	methods[6].fnPtr = (void*)Java_nvcmd_Int_1GetCommand;

	// GetInContainer function
	methods[7].name = (char *)"Int_GetInContainer";
	methods[7].signature = (char *)"(I)Ljava/lang/String;";
	methods[7].fnPtr = (void*)Java_nvcmd_Int_1GetInContainer;

	// GetOutContainer function
	methods[8].name = (char *)"Int_GetOutContainer";
	methods[8].signature = (char *)"(I)Ljava/lang/String;";
	methods[8].fnPtr = (void*)Java_nvcmd_Int_1GetOutContainer;

	// GetLanguage function
	methods[9].name = (char *)"Int_GetLanguage";
	methods[9].signature = (char *)"(I)Ljava/lang/String;";
	methods[9].fnPtr = (void*)Java_nvcmd_Int_1GetLanguage;

	// GetMachineName function
	methods[10].name = (char *)"Int_GetMachineName";
	methods[10].signature = (char *)"(I)Ljava/lang/String;";
	methods[10].fnPtr = (void*)Java_nvcmd_Int_1GetMachineName;

	// GetMachineUser function
	methods[11].name = (char *)"Int_GetMachineUser";
	methods[11].signature = (char *)"(I)Ljava/lang/String;";
	methods[11].fnPtr = (void*)Java_nvcmd_Int_1GetMachineUser;

	// GetError function
	methods[12].name = (char *)"Int_GetError";
	methods[12].signature = (char *)"(ILjava/lang/String;)Ljava/lang/String;";
	methods[12].fnPtr = (void*)Java_nvcmd_Int_1GetError;

	// GetNumParameters function
	methods[13].name = (char *)"Int_GetNumParameters";
	methods[13].signature = (char *)"(I)I";
	methods[13].fnPtr = (void*)Java_nvcmd_Int_1GetNumParameters;

	// GetParameterName function
	methods[14].name = (char *)"Int_GetParameterName";
	methods[14].signature = (char *)"(II)Ljava/lang/String;";
	methods[14].fnPtr = (void*)Java_nvcmd_Int_1GetParameterName;

	// GetSource function
	methods[15].name = (char *)"Int_GetSource";
	methods[15].signature = (char *)"(I)Ljava/lang/String;";
	methods[15].fnPtr = (void*)Java_nvcmd_Int_1GetSource;

	// IsCommand function
	methods[16].name = (char *)"Int_IsCommand";
	methods[16].signature = (char *)"(ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;)Z";
	methods[16].fnPtr = (void*)Java_nvcmd_Int_1IsCommand;

	// ParameterExist function
	methods[17].name = (char *)"Int_ParameterExist";
	methods[17].signature = (char *)"(ILjava/lang/String;)Z";
	methods[17].fnPtr = (void*)Java_nvcmd_Int_1ParameterExist;

	// GetParameter function
	methods[18].name = (char *)"Int_GetParameter";
	methods[18].signature = (char *)"(ILjava/lang/String;)Ljava/lang/String;";
	methods[18].fnPtr = (void*)Java_nvcmd_Int_1GetParameter;

	// AddHttpHeader function
	methods[19].name = (char *)"Int_AddHttpHeader";
	methods[19].signature = (char *)"(ILjava/lang/String;Ljava/lang/String;)V";
	methods[19].fnPtr = (void*)Java_nvcmd_Int_1AddHttpHeader;

	// Do the registration here
	i = env->RegisterNatives(nvcls4, methods, 20);

	// We don't need that any more
	delete methods;

	if(i != 0)
	{
		jvm->DestroyJavaVM();
		jvm = NULL;
#ifndef JVM_IMPLICIT
		MyFreeLibrary(hJvm);
		hJvm = NULL;
#endif
		return 0;
	}
*/
	// Everything is OK

	return 1;
}


// Unloads the JVM
// The load and unload of the JVM must be made from the same thread
void NvFreeJava()
{
	// Do nothing if not loaded
	if(jvm == NULL)
		return;

	JNIEnv *env;
	MyAttachCurrentThread((void **)(&env));
	if(env != NULL)
	{
		if (env->ExceptionOccurred())
			env->ExceptionClear();
	}

	// Remove the env list
	EnvList.RemoveAll();

	MyDetachCurrentThread();

	// Remove the thread list
	ThreadList.RemoveAll();
	
	// Destroy the JVM
//	jvm->ResetJavaVM();
	jvm->DestroyJavaVM();
	jvm = NULL;

	// Unload the library
#ifndef JVM_IMPLICIT
	MyFreeLibrary(hJvm);
	hJvm = NULL;
#endif
}


int NvRunMethod(const char* MethodName, void *pCommand)
{
	if(MethodName == NULL)
		return 0;
	if(pCommand == NULL)
		return 0;

	string DirName;
	string ClassName;
	string RealMethodName;

	if(!ParseMethodName(MethodName, DirName, ClassName, RealMethodName))
		return 0;

	// Now get our own thread context
	NvJavaInt* JavaInt = ThreadList.GetThreadNvIntPtr(pCommand);
	if(JavaInt == NULL)
	{
		MyDetachCurrentThread();
		return 0;
	}

	NvJavaEnv JavaEnv;
	JavaEnv.bOkOne = false;		// retval of 0 is OK from procedure
	if(!JavaEnv.LoadClass(DirName, ClassName))
	{
		ThreadList.FreeThreadNvIntPtr();
		return 0;
	}
	
	if(!JavaEnv.RunMethod(RealMethodName))
	{
		JavaEnv.UnLoadClass();
		ThreadList.FreeThreadNvIntPtr();
		return 0;
	}

	JavaEnv.UnLoadClass();

	// We must detach the thread (the real detach is made only if this is the last of recursive calls)
	ThreadList.FreeThreadNvIntPtr();

	return 1;
}




int ParseMethodName(const char* MethodNameToParse, string& DirName, string& ClassName, string& MethodName)
{
	if(MethodNameToParse == NULL)
		return 0;

	DirName = "";
	ClassName = "";
	MethodName = "";

	// We must first extract the class path, name and method to use. 
	MethodName = MethodNameToParse;
	CleanString(MethodName);
	SetDirName(MethodName);

	DirName = MethodName;
	ClassName = MethodName;
	string::size_type index;
	index = DirName.rfind('\\');
	if(index == string::npos)
		index = DirName.rfind('/');

	if(index == string::npos)
	{
		DirName = "";
	}
	else
	{
		ClassName = DirName.substr(index+1);
		DirName = DirName.substr(0, index);
		CleanString(DirName);
		CleanString(ClassName);
	}

	if(!ClassName.length())
		return 0;

	MethodName = "main";
	index = ClassName.find(':');
	if(index != string::npos)
	{
		MethodName = ClassName.substr(index+1);
		ClassName = ClassName.substr(0, index);
		CleanString(ClassName);
		CleanString(MethodName);
	}

	if(!ClassName.length())
		return 0;
	if(!MethodName.length())
		MethodName = "main";

	AddDirSuffix(DirName);

	// Now we have DirName (may be empty) , ClassName and real method name

	// Since the class name is case independant, we must try to get the real class name
	string ClassFileName = DirName;
	AddDirSuffix(DirName);
	ClassFileName += ClassName;
	ClassFileName += ".class";
	if(GetNoCaseRealFileName(ClassFileName))
	{
#ifdef WIN32
		index = ClassFileName.rfind('\\');
#else
		index = ClassFileName.rfind('/');
#endif
		string NewClassName = ClassName; 
		if(index != string::npos)
			NewClassName = ClassFileName.substr(index+1);
		index = NewClassName.rfind('.');
		if(index == string::npos)
			NewClassName = ClassName;
		else
			NewClassName = NewClassName.substr(0, index);
		ClassName = NewClassName;
	}
	else
	{
		// The file has not been found
		// We'll try to find a directory or jar file
		if(!DirName.length())
			return 0;			// nothing foundable in this case
		string ProviStr = DirName;
		// Remove last slash or backslash
		ProviStr = ProviStr.substr(0, ProviStr.length()-1);
		ProviStr += ".jar";
		if(GetNoCaseRealFileName(ProviStr))
		{
			// Yes, it's a jar file
			DirName = ProviStr;
		}
		// If not a jar file, we suppose that it's a directory
	}

	return 1;
}



// Detach from external
// Should be called by security when a thread is released
int NvDetachThread()
{
	// Detach in force mode
	ThreadList.FreeThreadNvIntPtr(true);

	return 1;
}



// Java interface methods
JNIEXPORT jint JNICALL Java_nvsint_Int_1GetCommandHandle(JNIEnv *env, jobject jobj)
{
	return (jint)Java_com_nirvasoft_nirva_nvsint_Int_1GetCommandHandle(env, jobj);
}
JNIEXPORT jlong JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetCommandHandle(JNIEnv *env, jobject jobj)
{
	return ((jlong)(ThreadList.GetThreadNvIntPtr()));
}


JNIEXPORT void JNICALL Java_nvsint_Int_1SetError(JNIEnv *env, jobject jobj, jint CommandHandle, jstring ErrorInfo)
{
	Java_com_nirvasoft_nirva_nvsint_Int_1SetError(env, jobj, (jlong)CommandHandle, ErrorInfo);
}
JNIEXPORT void JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1SetError(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring ErrorInfo)
{
	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return;

	try
	{
		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return;

		string ErrorInfoHere = "";
		if(ErrorInfo != NULL)
		{
			const char *str = env->GetStringUTFChars(ErrorInfo, 0);
			ErrorInfoHere = str;
			env->ReleaseStringUTFChars(ErrorInfo, str);
		}
		pCommand->SetError(ErrorInfoHere.c_str());
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (SetError) (SetError) ***\n");
	}
}

JNIEXPORT void JNICALL Java_nvsint_Int_1SetErrorEx(JNIEnv *env, jobject jobj, jint CommandHandle, jstring ErrorService, jstring ErrorClass, jint ErrorCode, jstring ErrorInfo)
{
	Java_com_nirvasoft_nirva_nvsint_Int_1SetErrorEx(env, jobj, (jlong)CommandHandle, ErrorService, ErrorClass, ErrorCode, ErrorInfo);
}
JNIEXPORT void JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1SetErrorEx(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring ErrorService, jstring ErrorClass, jint ErrorCode, jstring ErrorInfo)
{
	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return;

	try
	{
		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return;

		string ErrorServiceHere = "";
		if(ErrorService != NULL)
		{
			const char *str = env->GetStringUTFChars(ErrorService, 0);
			ErrorServiceHere = str;
			env->ReleaseStringUTFChars(ErrorService, str);
		}

		string ErrorClassHere = "";
		if(ErrorClass != NULL)
		{
			const char *str = env->GetStringUTFChars(ErrorClass, 0);
			ErrorClassHere = str;
			env->ReleaseStringUTFChars(ErrorClass, str);
		}

		string ErrorInfoHere = "";
		if(ErrorInfo != NULL)
		{
			const char *str = env->GetStringUTFChars(ErrorInfo, 0);
			ErrorInfoHere = str;
			env->ReleaseStringUTFChars(ErrorInfo, str);
		}

		pCommand->SetErrorEx(ErrorServiceHere.c_str(), ErrorClassHere.c_str(), ErrorCode, ErrorInfoHere.c_str());
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (SetErrorEx) ***\n");
	}
}


JNIEXPORT jint JNICALL Java_nvsint_Int_1Command(JNIEnv *env, jobject jobj, jint CommandHandle, jstring CommandString)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1Command(env, jobj, (jlong)CommandHandle, CommandString);
}
JNIEXPORT jint JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1Command(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring CommandString)
{
	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return 0;

	try
	{
		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return 0;

		string CommandStringHere = "";
		if(CommandString != NULL)
		{
			const char *str = env->GetStringUTFChars(CommandString, 0);
			CommandStringHere = str;
			env->ReleaseStringUTFChars(CommandString, str);
		}

		strcpy(JavaInt->ResultBuffer, "");
		int DataSize;
		if(!pCommand->Command(CommandStringHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize))
			return 0;
		int ResultSetBufferSize = JavaInt->SetBufferSize(DataSize);
		if(ResultSetBufferSize == 0)
			return 0;	// Error
		if(ResultSetBufferSize == 2)
		{
			// We got a reallocation so we must get the last output buffer
			pCommand->Command("NV_CMD=|COMMAND:GET_LAST_OUTPUT_BUFFER|", JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
//			if(!pCommand->Command(CommandStringHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize))
//				return 0;
		}

		JavaInt->LastResult = JavaInt->ResultBuffer;

	}
	catch(...)
	{
		printf("*** Exception occured in java interface (Command) ***\n");
	}
	return 1;
}

JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetResult(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetResult(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetResult(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;

	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		Result = JavaInt->LastResult;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetResult) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetSessionId(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetSessionId(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetSessionId(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetSessionId(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetSessionId(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetSessionId) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}


JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetService(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetService(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetService(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetService(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetService(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetService) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}


JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetClass(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetClass(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetClass(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetClass(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetClass(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetClass) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetCommand(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetCommand(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetCommand(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetCommand(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetCommand(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetCommand) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetInContainer(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetInContainer(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetInContainer(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetInContainer(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetInContainer(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetInContainer) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetOutContainer(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetOutContainer(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetOutContainer(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetOutContainer(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetOutContainer(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetOutContainer) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetLanguage(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetLanguage(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetLanguage(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetLanguage(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetLanguage(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetLanguage) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetError(JNIEnv *env, jobject jobj, jint CommandHandle, jstring InfoType)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetError(env, jobj, (jlong)CommandHandle, InfoType);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetError(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring InfoType)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());


		string InfoTypeHere = "";
		if(InfoType != NULL)
		{
			const char *str = env->GetStringUTFChars(InfoType, 0);
			InfoTypeHere = str;
			env->ReleaseStringUTFChars(InfoType, str);
		}

		int DataSize;
		pCommand->GetError(InfoTypeHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetError(InfoTypeHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetError) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jint JNICALL Java_nvsint_Int_1GetNumParameters(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetNumParameters(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jint JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetNumParameters(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	int Result = 0;

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return Result;

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return Result;

		Result = pCommand->GetNumParameters();
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetNumParameters) ***\n");
	}

	return Result;
}


JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetSource(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetSource(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetSource(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "UNKNOWN";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int SourceType = pCommand->GetSource();
		switch(SourceType)
		{
			case NV_SOURCE_CLIENT:
				Result = "CLIENT";
				break;
			case NV_SOURCE_BROWSER:
				Result = "BROWSER";
				break;
			case NV_SOURCE_PROCEDURE:
				Result = "PROCEDURE";
				break;
			case NV_SOURCE_SERVICE:
				Result = "SERVICE";
				break;
			case NV_SOURCE_SERVER:
				Result = "SERVER";
				break;
		}
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetSource) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}



JNIEXPORT jint JNICALL Java_nvsint_Int_1ParameterExist(JNIEnv *env, jobject jobj, jint CommandHandle, jstring ParameterName)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1ParameterExist(env, jobj, (jlong)CommandHandle, ParameterName);
}
JNIEXPORT jint JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1ParameterExist(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring ParameterName)
{
	int Result = 0;

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return Result;

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return Result;


		string ParameterNameHere = "";
		if(ParameterName != NULL)
		{
			const char *str = env->GetStringUTFChars(ParameterName, 0);
			ParameterNameHere = str;
			env->ReleaseStringUTFChars(ParameterName, str);
		}

		if(pCommand->ParameterExist(ParameterNameHere.c_str()))
			Result = 1;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (ParameterExist) ***\n");
	}
	return Result;
}



JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetParameter(JNIEnv *env, jobject jobj, jint CommandHandle, jstring ParameterName)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetParameter(env, jobj, (jlong)CommandHandle, ParameterName);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetParameter(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring ParameterName)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());


		string ParameterNameHere = "";
		if(ParameterName != NULL)
		{
			const char *str = env->GetStringUTFChars(ParameterName, 0);
			ParameterNameHere = str;
			env->ReleaseStringUTFChars(ParameterName, str);
		}

		int DataSize;
		pCommand->GetParameter(ParameterNameHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetParameter(ParameterNameHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetParameter) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}




JNIEXPORT jstring JNICALL Java_nvsint_Int_1GetParameterName(JNIEnv *env, jobject jobj, jint CommandHandle, jint index)
{
	return Java_com_nirvasoft_nirva_nvsint_Int_1GetParameterName(env, jobj, (jlong)CommandHandle, index);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvsint_Int_1GetParameterName(JNIEnv *env, jobject jobj, jlong CommandHandle, jint index)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvJavaCommand *pCommand = NULL;
	try
	{
		if(JavaInt == NULL)
			return env->NewStringUTF(Result.c_str());

		pCommand = (NvJavaCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());


		int DataSize;
		pCommand->GetParameterName(index, JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetParameterName(index, JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetParameterName) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}


