// ***************************************************************************************
// File: service.cpp
// Description: Nirva interface to java service
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifdef WIN32
#pragma warning( disable : 4996 )
#endif

#include "task.h"
#include "util.h"
#include "nvcmd.h"
#include "ejava.h"
#include "global.h"
#include "service.h"
#include <map>
#include <string>

using namespace std;


extern JavaVM *jvm;
extern NvJavaThreadList ThreadList;
extern NvJavaEnvList EnvList;


int NvInitService(NvServiceCommand *Command, string& ServiceName);			// Service init
int NvExitService(NvServiceCommand *Command, string& ServiceName);			// Service exit
int NvCommand(NvServiceCommand *Command, string& ServiceName);				// Real command
int NvRender(NvRendererInt *Render, string& ServiceName);					// Service renderer

// Main Nirva external service entry point
// This function should not be modified
// All implementation must be made in the NVJAVASession class
int NvsCommand(void *pCommand)
{
	// The pCommand points in fact on a NvServiceCommand class
	NvServiceCommand *Command = (NvServiceCommand *)pCommand;
	int RetVal = 0;

	// We get the service name
	int BufferSize;
	Command->GetParameter("NV_JAVA_SERVICE_NAME", NULL, 0, &BufferSize);
	BufferSize += 10;
	char *Buffer = (char *)malloc(BufferSize);
	Command->GetParameter("NV_JAVA_SERVICE_NAME", Buffer, BufferSize);
	string ServiceName = Buffer;
	free(Buffer);

//	string CommandToCheck = "NV_INIT_SERVICE#&NV&#";
//	CommandToCheck += ServiceName;

	// Check the service init and exit commands
	if(Command->IsCommand("SYSTEM", "NV_INIT_SERVICE", NV_SOURCE_SERVER))
		return NvInitService(Command, ServiceName);

//	CommandToCheck = "NV_EXIT_SERVICE#&NV&#";
//	CommandToCheck += ServiceName;

	if(Command->IsCommand("SYSTEM", "NV_EXIT_SERVICE", NV_SOURCE_SERVER))
		return NvExitService(Command, ServiceName);

	// Normal command
	RetVal = NvCommand(Command, ServiceName);

	return RetVal;
}



// Called one time when the service is first acceded
int NvInitService(NvServiceCommand *Command, string& ServiceName)
{
	char Buffer[1024];
	Command->GetParameter("NV_JAVA_SERVICE_PATH", Buffer, sizeof(Buffer));
	string ServicePath = Buffer;
	if(!ServicePath.length())
		return 0;
	string DirName;
	string ClassName;
	string MethodName;

	if(!ParseMethodName(ServicePath.c_str(), DirName, ClassName, MethodName))
		return 0;
	
	// Always Init here
	MethodName = "Init";

	// Now get our own thread context
	NvJavaInt* JavaInt = ThreadList.GetThreadNvIntPtr(Command);
	if(JavaInt == NULL)
	{
		MyDetachCurrentThread();
		return 0;
	}

	NvJavaEnv* JavaEnv = EnvList.GetJavaEnv(ServiceName);
	if(JavaEnv != NULL)
	{
		ThreadList.FreeThreadNvIntPtr();
		return 1;	// ever initalized
	}

	JavaEnv = new NvJavaEnv;

	if(!JavaEnv->LoadClass(DirName, ClassName))
	{
		delete JavaEnv;
		ThreadList.FreeThreadNvIntPtr();
		return 0;
	}
	
	if(!JavaEnv->RunMethod(MethodName, true))
	{
		delete JavaEnv;
		ThreadList.FreeThreadNvIntPtr();
		return 0;
	}

	EnvList.AddJavaEnv(ServiceName, JavaEnv);

	// We must detach the thread (the real detach is made only if this is the last of recursive calls)
	ThreadList.FreeThreadNvIntPtr();

	return 1;
}


// Called one time when the service is closed
int NvExitService(NvServiceCommand *Command, string& ServiceName)
{
	// Always Command here
	string MethodName = "Exit";


	// Now get our own thread context
	NvJavaInt* JavaInt = ThreadList.GetThreadNvIntPtr(Command);
	if(JavaInt == NULL)
	{
		MyDetachCurrentThread();
		return 0;
	}

	NvJavaEnv* JavaEnv = EnvList.GetJavaEnv(ServiceName);
	if(JavaEnv == NULL)
	{
		ThreadList.FreeThreadNvIntPtr();
		return 0;	// not initialized, should not occur
	}


	if(!JavaEnv->RunMethod(MethodName, true))
	{
		ThreadList.FreeThreadNvIntPtr();
		return 0;
	}

	EnvList.RemoveJavaEnv(ServiceName);

	// We must detach the thread (the real detach is made only if this is the last of recursive calls)
	ThreadList.FreeThreadNvIntPtr();

	return 1;
}


// Renderer entry point
int NvsRender(void *RenderInterface, const char *ServiceName)
{
	// The RenderInterface points in fact on a NvRendererInt class
	NvRendererInt *Render = (NvRendererInt*)RenderInterface;
	int RetVal = 0;
	string ServiceNameHere = ServiceName;

	// Do the job
	RetVal = NvRender(Render, ServiceNameHere);

	return RetVal;
}


// Called for each command
int NvCommand(NvServiceCommand *Command, string& ServiceName)
{
	
	// Always Command here
	string MethodName = "Command";

	// Now get our own thread context
	NvJavaEnv* JavaEnv = EnvList.GetJavaEnv(ServiceName);
	NvJavaInt* JavaInt = ThreadList.GetThreadNvIntPtr(Command);
	if(JavaInt == NULL)
	{
		MyDetachCurrentThread();
		return 0;
	}

	if(JavaEnv == NULL)
	{
		ThreadList.FreeThreadNvIntPtr();
		return 0;	// not initialized, should not occur
	}


	if(!JavaEnv->RunMethod(MethodName, true))
	{
		ThreadList.FreeThreadNvIntPtr();
		return 0;
	}

	// We must detach the thread (the real detach is made only if this is the last of recursive calls)
	ThreadList.FreeThreadNvIntPtr();

	return 1;
}


// Called for each render request
// The renderer is called after any command so there is no need to maitain a command stack
// The thread can be attached to java before rendering and detached after rendering
int NvRender(NvRendererInt *Render, string& ServiceName)
{
	// Get the java service environment
	NvJavaEnv* JavaEnv = EnvList.GetJavaEnv(ServiceName);

	if(JavaEnv == NULL)
	{
		Render->SetReason("Java service not initialized");
		return 0;	// not initialized, should not occur
	}

	if(!JavaEnv->DoRender(Render))
	{
		MyDetachCurrentThread();
		return 0;
	}
	
	// Detach the thread
	MyDetachCurrentThread();

	return 1;
}



// Java interface methods
JNIEXPORT jint JNICALL Java_nvcmd_Int_1GetCommandHandle(JNIEnv *env, jobject jobj)
{
	return (jint)Java_com_nirvasoft_nirva_nvcmd_Int_1GetCommandHandle(env, jobj);
}
JNIEXPORT jlong JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetCommandHandle(JNIEnv *env, jobject jobj)
{
	return ((jlong)(ThreadList.GetThreadNvIntPtr()));
}

JNIEXPORT void JNICALL Java_nvcmd_Int_1SetError(JNIEnv *env, jobject jobj, jint CommandHandle, jstring ErrorClass, jint ErrorCode, jstring ErrorInfo)
{
	Java_com_nirvasoft_nirva_nvcmd_Int_1SetError(env, jobj, (jlong)CommandHandle, ErrorClass, ErrorCode, ErrorInfo);
}
JNIEXPORT void JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1SetError(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring ErrorClass, jint ErrorCode, jstring ErrorInfo)
{
	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return;

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return;

		string ErrorClassHere = "";
		if(ErrorClass != NULL)
		{
			const char *str = env->GetStringUTFChars(ErrorClass, 0);
			ErrorClassHere = str;
			env->ReleaseStringUTFChars(ErrorClass, str);
		}

		string ErrorInfoHere = "";
		if(ErrorInfo != NULL)
		{
			const char *str = env->GetStringUTFChars(ErrorInfo, 0);
			ErrorInfoHere = str;
			env->ReleaseStringUTFChars(ErrorInfo, str);
		}

		pCommand->SetError(ErrorClassHere.c_str(), ErrorCode, ErrorInfoHere.c_str());
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (SetError) ***\n");
	}
}


JNIEXPORT jboolean JNICALL Java_nvcmd_Int_1Command(JNIEnv *env, jobject jobj, jint CommandHandle, jstring CommandString)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1Command(env, jobj, (jlong)CommandHandle, CommandString);
}
JNIEXPORT jboolean JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1Command(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring CommandString)
{
	jboolean Result = JNI_FALSE;
	NvServiceCommand *pCommand = NULL;
	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	if(JavaInt == NULL)
		return JNI_FALSE;

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return JNI_FALSE;


		string CommandStringHere = "";
		if(CommandString != NULL)
		{
			const char *str = env->GetStringUTFChars(CommandString, 0);
			CommandStringHere = str;
			env->ReleaseStringUTFChars(CommandString, str);
		}

		strcpy(JavaInt->ResultBuffer, "");
		int DataSize;
		if(!pCommand->Command(CommandStringHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize))
			return JNI_FALSE;
		Result = JNI_TRUE;
		int ResultSetBufferSize = JavaInt->SetBufferSize(DataSize);
		if(ResultSetBufferSize == 0)
			return JNI_FALSE;	// Error
		if(ResultSetBufferSize == 2)
		{
			// We got a reallocation so we must get the last output buffer
			pCommand->Command("NV_CMD=|COMMAND:GET_LAST_OUTPUT_BUFFER|", JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
//			if(!pCommand->Command(CommandStringHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize))
//				return JNI_FALSE;
//			Result = JNI_TRUE;
		}

		JavaInt->LastResult = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (Command) ***\n");
		Result = JNI_FALSE;
	}

	return Result;
}

JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetResult(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetResult(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetResult(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		Result = JavaInt->LastResult;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetResult) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetSessionId(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetSessionId(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetSessionId(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetSessionId(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetSessionId(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetSessionId) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetClass(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetClass(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetClass(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetClass(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetClass(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetClass) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetCommand(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetCommand(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetCommand(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetCommand(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetCommand(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetCommand) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetInContainer(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetInContainer(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetInContainer(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetInContainer(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetInContainer(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetInContainer) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetOutContainer(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetOutContainer(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetOutContainer(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetOutContainer(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetOutContainer(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetOutContainer) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetLanguage(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetLanguage(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetLanguage(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetLanguage(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetLanguage(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetLanguage) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetMachineName(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetMachineName(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetMachineName(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetMachineName(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetMachineName(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetMachineName) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetMachineUser(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetMachineUser(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetMachineUser(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int DataSize;
		pCommand->GetMachineUser(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetMachineUser(JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetMachineUser) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetError(JNIEnv *env, jobject jobj, jint CommandHandle, jstring InfoType)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetError(env, jobj, (jlong)CommandHandle, InfoType);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetError(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring InfoType)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());


		string InfoTypeHere = "";
		if(InfoType != NULL)
		{
			const char *str = env->GetStringUTFChars(InfoType, 0);
			InfoTypeHere = str;
			env->ReleaseStringUTFChars(InfoType, str);
		}

		int DataSize;
		pCommand->GetError(InfoTypeHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetError(InfoTypeHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetError) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}

JNIEXPORT jint JNICALL Java_nvcmd_Int_1GetNumParameters(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetNumParameters(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jint JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetNumParameters(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	int Result = 0;

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return Result;

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return Result;

		Result = pCommand->GetNumParameters();

	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetNumParameters) ***\n");
	}
	return Result;
}


JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetSource(JNIEnv *env, jobject jobj, jint CommandHandle)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetSource(env, jobj, (jlong)CommandHandle);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetSource(JNIEnv *env, jobject jobj, jlong CommandHandle)
{
	string Result = "UNKNOWN";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());

		int SourceType = pCommand->GetSource();
		switch(SourceType)
		{
			case NV_SOURCE_CLIENT:
				Result = "CLIENT";
				break;
			case NV_SOURCE_BROWSER:
				Result = "BROWSER";
				break;
			case NV_SOURCE_PROCEDURE:
				Result = "PROCEDURE";
				break;
			case NV_SOURCE_SERVICE:
				Result = "SERVICE";
				break;
			case NV_SOURCE_SERVER:
				Result = "SERVER";
				break;
		}
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetSource) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}



JNIEXPORT jboolean JNICALL Java_nvcmd_Int_1ParameterExist(JNIEnv *env, jobject jobj, jint CommandHandle, jstring ParameterName)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1ParameterExist(env, jobj, (jlong)CommandHandle, ParameterName);
}
JNIEXPORT jboolean JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1ParameterExist(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring ParameterName)
{
	jboolean Result = JNI_FALSE;

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return Result;

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return Result;


		string ParameterNameHere = "";
		if(ParameterName != NULL)
		{
			const char *str = env->GetStringUTFChars(ParameterName, 0);
			ParameterNameHere = str;
			env->ReleaseStringUTFChars(ParameterName, str);
		}

		if(pCommand->ParameterExist(ParameterNameHere.c_str()))
			Result = JNI_TRUE;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (ParameterExist) ***\n");
	}
	return Result;
}




JNIEXPORT jboolean JNICALL Java_nvcmd_Int_1IsCommand(JNIEnv *env, jobject jobj, jint CommandHandle, jstring Class, jstring Command, jstring Source)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1IsCommand(env, jobj, (jlong)CommandHandle, Class, Command, Source);
}
JNIEXPORT jboolean JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1IsCommand(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring Class, jstring Command, jstring Source)
{
	jboolean Result = JNI_FALSE;

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return Result;

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return Result;


		string ClassHere = "";
		if(Class != NULL)
		{
			const char *str = env->GetStringUTFChars(Class, 0);
			ClassHere = str;
			CleanString(ClassHere);
			env->ReleaseStringUTFChars(Class, str);
		}

		string CommandHere = "";
		if(Command != NULL)
		{
			const char *str = env->GetStringUTFChars(Command, 0);
			CommandHere = str;
			CleanString(CommandHere);
			env->ReleaseStringUTFChars(Command, str);
		}

		string SourceHere = "";
		if(Source != NULL)
		{
			const char *str = env->GetStringUTFChars(Source, 0);
			SourceHere = str;
			MakeUpper(SourceHere);
			CleanString(SourceHere);
			env->ReleaseStringUTFChars(Source, str);
		}

		const char *pClassHere = NULL;
		const char *pCommandHere = NULL;
		int iSource = -1;
		if(SourceHere == "CLIENT")
			iSource = NV_SOURCE_CLIENT;
		if(SourceHere == "BROWSER")
			iSource = NV_SOURCE_BROWSER;
		if(SourceHere == "PROCEDURE")
			iSource = NV_SOURCE_PROCEDURE;
		if(SourceHere == "SERVICE")
			iSource = NV_SOURCE_SERVICE;
		if(SourceHere == "SERVER")
			iSource = NV_SOURCE_SERVER;

		if(ClassHere.length())
			pClassHere = ClassHere.c_str();

		if(CommandHere.length())
			pCommandHere = CommandHere.c_str();

		if(pCommand->IsCommand(pClassHere, pCommandHere, iSource))
			Result = JNI_TRUE;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (IsCommand) ***\n");
		Result = JNI_FALSE;
	}
	return Result;
}


JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetParameter(JNIEnv *env, jobject jobj, jint CommandHandle, jstring ParameterName)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetParameter(env, jobj, (jlong)CommandHandle, ParameterName);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetParameter(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring ParameterName)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());


		string ParameterNameHere = "";
		if(ParameterName != NULL)
		{
			const char *str = env->GetStringUTFChars(ParameterName, 0);
			ParameterNameHere = str;
			env->ReleaseStringUTFChars(ParameterName, str);
		}

		int DataSize;
		pCommand->GetParameter(ParameterNameHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetParameter(ParameterNameHere.c_str(), JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetParameter) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}




JNIEXPORT jstring JNICALL Java_nvcmd_Int_1GetParameterName(JNIEnv *env, jobject jobj, jint CommandHandle, jint index)
{
	return Java_com_nirvasoft_nirva_nvcmd_Int_1GetParameterName(env, jobj, (jlong)CommandHandle, index);
}
JNIEXPORT jstring JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1GetParameterName(JNIEnv *env, jobject jobj, jlong CommandHandle, jint index)
{
	string Result = "";

	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return env->NewStringUTF(Result.c_str());

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return env->NewStringUTF(Result.c_str());


		int DataSize;
		pCommand->GetParameterName(index, JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);
		if(JavaInt->SetBufferSize(DataSize) == 2)
			pCommand->GetParameterName(index, JavaInt->ResultBuffer, JavaInt->ResultBufferSize, &DataSize);

		Result = JavaInt->ResultBuffer;
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (GetParameterName) ***\n");
		Result = "";
	}
	return env->NewStringUTF(Result.c_str());
}



JNIEXPORT void JNICALL Java_nvcmd_Int_1AddHttpHeader(JNIEnv *env, jobject jobj, jint CommandHandle, jstring HeaderName, jstring HeaderValue)
{
	Java_com_nirvasoft_nirva_nvcmd_Int_1AddHttpHeader(env, jobj, (jlong)CommandHandle, HeaderName, HeaderValue);
}
JNIEXPORT void JNICALL Java_com_nirvasoft_nirva_nvcmd_Int_1AddHttpHeader(JNIEnv *env, jobject jobj, jlong CommandHandle, jstring HeaderName, jstring HeaderValue)
{
	NvJavaInt* JavaInt = (NvJavaInt*)CommandHandle;
	NvServiceCommand *pCommand = NULL;
	if(JavaInt == NULL)
		return;

	try
	{
		pCommand = (NvServiceCommand *)(JavaInt->m_pCommand);
		if(pCommand == NULL)
			return;

		string HeaderNameHere = "";
		if(HeaderName != NULL)
		{
			const char *str = env->GetStringUTFChars(HeaderName, 0);
			HeaderNameHere = str;
			env->ReleaseStringUTFChars(HeaderName, str);
		}

		string HeaderValueHere = "";
		if(HeaderValue != NULL)
		{
			const char *str = env->GetStringUTFChars(HeaderValue, 0);
			HeaderValueHere = str;
			env->ReleaseStringUTFChars(HeaderValue, str);
		}

		pCommand->AddHttpHeader(HeaderNameHere.c_str(), HeaderValueHere.c_str());
	}
	catch(...)
	{
		printf("*** Exception occured in java interface (AddHttpHeader) ***\n");
	}
}




