//*****************************************************************************
// File: application.cpp
// Description: Nirva application objects
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER


using namespace std;



/////////////////////////////////////////////////////////////////////////////
// NvApplicationList

NvApplicationList::NvApplicationList()
{
	// We create the default application
}

NvApplicationList::~NvApplicationList()
{
	Clear();
}



void NvApplicationList::Clear(bool AlsoDefault)
{
	int i;
	NV_LOCKOBJECT(this);
	NvApplication *DefaultApplication = NULL;
	for(i=0; i<(int)(m_vApplications.size()); i++)
	{
		NvApplication *Application = (NvApplication *)m_vApplications[i];
		if(Application != NULL)
		{
			if(AlsoDefault)
			{
				delete Application;
			}
			else
			{
				if(Application->m_sName == NV_APPLICATION_DEFAULT_NAME)
					DefaultApplication = Application;
				else
					delete Application;
			}
		}
	}
	m_vApplications.clear();
	if(DefaultApplication != NULL)
		m_vApplications.push_back((void *)DefaultApplication);
	NV_UNLOCKOBJECT(this);
}


void NvApplicationList::ClearRegistryCache()
{
	int i;
	NV_LOCKOBJECT(this);
	for(i=0; i<(int)(m_vApplications.size()); i++)
	{
		NvApplication *Application = (NvApplication *)m_vApplications[i];
		if(Application != NULL)
			if(!Application->bInInstall)
				Application->AppRegistry.Clear();
	}
	NV_UNLOCKOBJECT(this);
}


// Clears the application work directories
void NvApplicationList::CleanWorkDir()
{
	int i;
	NV_LOCKOBJECT(this);
	for(i=0; i<(int)(m_vApplications.size()); i++)
	{
		NvApplication *Application = (NvApplication *)m_vApplications[i];
		if(Application != NULL)
			if(!Application->bInInstall)
				RemoveDir(Application->GetWorkDir().c_str(), false);			
	}
	NV_UNLOCKOBJECT(this);
}

// Load
// Loads the application list from the system registry
// This function is called only one time when starting the server so there is no need of locking
void NvApplicationList::Load(bool NvdefOnly)
{
	// load the application table
	// The application table is stored in a table object named "applications" in the
	// system registry key system.applications
	NvContainer Container;
	if(GetMain()->SystemRegistry.GetKey("system.applications", &Container, NV_CONTAINER_COPY_NO_APPEND, "applications", false, true))
	{
		NvObj *Object = Container.GetObject("applications");
		if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_TABLE))
		{
			NvObjTable *ApplicationTable = (NvObjTable*)Object;
			int index_Name = ApplicationTable->GetColumnIndex("Name");
			int index_Desc = ApplicationTable->GetColumnIndex("Description");
			int index_Enable = ApplicationTable->GetColumnIndex("Enable");
			int index_TimeOut = ApplicationTable->GetColumnIndex("TimeOut");
			int index_AuthMode = ApplicationTable->GetColumnIndex("AuthMode");
			int index_NirvaIsSsoUser = ApplicationTable->GetColumnIndex("NirvaIsSsoUser");
			int index_SecuApp = ApplicationTable->GetColumnIndex("SecuApp");
			int index_UseScheduled = ApplicationTable->GetColumnIndex("UseScheduled");
			int index_AllowWebXml = ApplicationTable->GetColumnIndex("AllowWebXml");
			int index_AllowNotOwnedSession = ApplicationTable->GetColumnIndex("AllowNotOwnedSession");
			int row_index;
			for(row_index=0; row_index<ApplicationTable->GetNumRow(); row_index++)
			{
				// Get application parameters
				string ApplicationName = ApplicationTable->GetCellLine(row_index, index_Name);
				NvApplication::CleanAppName(ApplicationName);
				if(!ApplicationName.length()) // The application is identified by its name
					continue;
				if(NvdefOnly && (ApplicationName != "NVDEF"))
					continue;
				if(InternalGetApplication(ApplicationName.c_str()) != NULL)	// Check doubles
					continue;
				string ApplicationDesc = ApplicationTable->GetCellLine(row_index, index_Desc);
				string ApplicationEnable = ApplicationTable->GetCellLine(row_index, index_Enable);
				bool bApplicationEnable = true;
				if(ApplicationEnable.length() && (atoi(ApplicationEnable.c_str()) == 0))
					bApplicationEnable = false;
				
				int DefaultTimeOut = 300;
				if(index_TimeOut != -1)
				{
					string sDefaultTimeOut = ApplicationTable->GetCellLine(row_index, index_TimeOut);
					DefaultTimeOut = atoi(sDefaultTimeOut.c_str());
					if(DefaultTimeOut < 0)
						DefaultTimeOut = 3600;
					if(DefaultTimeOut < 5)
						DefaultTimeOut = 5;
				}
				int AuthMode = NV_AUTHENTICATION_NIRVA;
				if(index_AuthMode != -1)
				{
					string sAuthMode = ApplicationTable->GetCellLine(row_index, index_AuthMode);
					AuthMode = atoi(sAuthMode.c_str());
					if(AuthMode < 0)
						AuthMode = NV_AUTHENTICATION_NIRVA;
					if(AuthMode > NV_AUTHENTICATION_MAXVALUE)
						AuthMode = NV_AUTHENTICATION_NIRVA;
				}

				bool NirvaIsSsoUser = false;
				if(index_NirvaIsSsoUser != -1)
				{
					string sNirvaIsSsoUser = ApplicationTable->GetCellLine(row_index, index_NirvaIsSsoUser);
					MakeUpper(sNirvaIsSsoUser);
					CleanString(sNirvaIsSsoUser);
					if(sNirvaIsSsoUser == "1")
						NirvaIsSsoUser = true;
				}

				string ApplicationSecuApp = "";
				if(index_SecuApp != -1)
					ApplicationSecuApp = ApplicationTable->GetCellLine(row_index, index_SecuApp);

				bool UseScheduled = true;
				if(index_UseScheduled != -1)
				{
					string sUseScheduled = ApplicationTable->GetCellLine(row_index, index_UseScheduled);
					MakeUpper(sUseScheduled);
					CleanString(sUseScheduled);
					if(sUseScheduled == "0")
						UseScheduled = false;
				}

				bool AllowWebXml = true;
				if(index_AllowWebXml != -1)
				{
					string sAllowWebXml = ApplicationTable->GetCellLine(row_index, index_AllowWebXml);
					MakeUpper(sAllowWebXml);
					CleanString(sAllowWebXml);
					if(sAllowWebXml == "0")
						AllowWebXml = false;
				}
				
				bool AllowNotOwnedSession = true;
				if(index_AllowNotOwnedSession != -1)
				{
					string sAllowNotOwnedSession = ApplicationTable->GetCellLine(row_index, index_AllowNotOwnedSession);
					MakeUpper(sAllowNotOwnedSession);
					CleanString(sAllowNotOwnedSession);
					if(sAllowNotOwnedSession == "0")
						AllowNotOwnedSession = false;
				}
				
				// Create the application in application list
				if(!CreateApplication(ApplicationName.c_str(), ApplicationDesc.c_str(), DefaultTimeOut, ApplicationSecuApp.c_str(), UseScheduled, AllowWebXml, true, AllowNotOwnedSession, false, AuthMode, NirvaIsSsoUser))
					bApplicationEnable = false;

				if(bApplicationEnable)
					EnableApplication(ApplicationName.c_str(), true);
			}
		}
	}

	// Saves the list in case of changes
	if(!NvdefOnly)
		Save();
}


// Saves the application list to the system registry
// This function is called internally so its not locked
bool NvApplicationList::Save()
{
	// First create the registry key "system.applications" if it doesn't exist
	GetMain()->SystemRegistry.SetKey("system.applications", NULL);

	NvContainer Container;
	if(!GetMain()->SystemRegistry.GetKey("system.applications", &Container, NV_CONTAINER_COPY_NO_APPEND, NULL, false, true))
		return false; // cannot get the key. This should not occur here

	// Creates the application object if it doesn't exist
	Container.CreateObject("applications", NV_OBJTYPE_TABLE);
	NvObj *Object = Container.GetObject("applications");
	if(Object == NULL)
		return false; // This should not occur here

	if(Object->GetType() != NV_OBJTYPE_TABLE)
	{
		Container.RemoveObject("applications");
		Container.CreateObject("applications", NV_OBJTYPE_TABLE);
		Object = Container.GetObject("applications");
		if(Object == NULL)
			return false; // This should not occur here
	}

	NvObjTable *ApplicationTable = (NvObjTable*)Object;
	ApplicationTable->Clear();
	ApplicationTable->InsertColumn("Name");
	ApplicationTable->InsertColumn("Description");
	ApplicationTable->InsertColumn("Enable");
	ApplicationTable->InsertColumn("TimeOut");
	ApplicationTable->InsertColumn("SecuApp");
	ApplicationTable->InsertColumn("UseScheduled");
	ApplicationTable->InsertColumn("AllowWebXml");
	ApplicationTable->InsertColumn("AllowNotOwnedSession");
	ApplicationTable->InsertColumn("AuthMode");
	ApplicationTable->InsertColumn("NirvaIsSsoUser");

	int index_Name = ApplicationTable->GetColumnIndex("Name");
	int index_Desc = ApplicationTable->GetColumnIndex("Description");
	int index_Enable = ApplicationTable->GetColumnIndex("Enable");
	int index_TimeOut = ApplicationTable->GetColumnIndex("TimeOut");
	int index_AuthMode = ApplicationTable->GetColumnIndex("AuthMode");
	int index_NirvaIsSsoUser = ApplicationTable->GetColumnIndex("NirvaIsSsoUser");
	int index_SecuApp = ApplicationTable->GetColumnIndex("SecuApp");
	int index_UseScheduled = ApplicationTable->GetColumnIndex("UseScheduled");
	int index_AllowWebXml = ApplicationTable->GetColumnIndex("AllowWebXml");
	int index_AllowNotOwnedSession = ApplicationTable->GetColumnIndex("AllowNotOwnedSession");

	int i,j;
	for(i=0, j=0; i<(int)(m_vApplications.size()); i++)
	{
		NvApplication *Application = (NvApplication *)m_vApplications[i];
		if(Application == NULL)
			continue;

		if(!Application->GetName().length())
			continue;

		ApplicationTable->InsertRow();
		ApplicationTable->InsertCellLine(j, index_Name, Application->GetName().c_str());
		ApplicationTable->InsertCellLine(j, index_Desc, Application->GetDesc().c_str());
		if(Application->IsEnable())
			ApplicationTable->InsertCellLine(j, index_Enable, "1");
		else
			ApplicationTable->InsertCellLine(j, index_Enable, "0");
		char Buf[20];
		sprintf(Buf, "%d", Application->GetDefaultTimeOut());
		ApplicationTable->InsertCellLine(j, index_TimeOut, Buf);
		sprintf(Buf, "%d", Application->GetAuthMode());
		ApplicationTable->InsertCellLine(j, index_AuthMode, Buf);
		strcpy(Buf, "0");
		if(Application->GetNirvaIsSsoUser())
			strcpy(Buf, "1");
		ApplicationTable->InsertCellLine(j, index_NirvaIsSsoUser, Buf);
		ApplicationTable->InsertCellLine(j, index_SecuApp, Application->GetSecuApp().c_str());
		strcpy(Buf, "1");
		if(!Application->GetUseScheduledTransac())
			strcpy(Buf, "0");
		ApplicationTable->InsertCellLine(j, index_UseScheduled, Buf);
		strcpy(Buf, "1");
		if(!Application->GetAllowWebXml())
			strcpy(Buf, "0");
		ApplicationTable->InsertCellLine(j, index_AllowWebXml, Buf);
		strcpy(Buf, "1");
		if(!Application->GetAllowNotOwnedSession())
			strcpy(Buf, "0");
		ApplicationTable->InsertCellLine(j, index_AllowNotOwnedSession, Buf);
		j++;
	}
	
	if(!GetMain()->SystemRegistry.SetKey("system.applications", &Container))
		return false;

	return true;
}


void NvApplicationList::Start()
{
	// Start all authorized applications
	// The NVDEF application must be started separatly (directly in nvs.cpp)
	// Not locked because called only from the nvs.cpp at initialisation time
	NV_LOCKLOCK(&ProtectStartStop);
	int i;
	// Get the list of authorized applications
	string Value;
	map<NvApplication*,int> AuthorizedApplications;
/*	if(GetMain()->Licenses.CheckKey("SYSTEM",NVS_SERVICE_ID,"APPLICATIONS", Value, false))
	{
		// The license is application limited
		// Check if time limited
		if(GetMain()->Licenses.CheckKey("SYSTEM",NVS_SERVICE_ID,"APPLICATIONS", Value))
		{
			// We are not time limited so we can run the authorized applications
			string AppName = "";
			string::size_type index;
			while(1)
			{
				index = Value.find(';');
				if(index != string::npos)
				{
					AppName = Value.substr(0, index);
					Value = Value.substr(index+1);
				}
				else
				{
					AppName = Value;
					Value = "";
				}
				NvApplication::CleanAppName(AppName);
				NvApplication *Application = InternalGetApplication(AppName.c_str());
				if((Application != NULL) && (Application->GetName() != NV_APPLICATION_DEFAULT_NAME))
					AuthorizedApplications[Application]=1;
				CleanString(Value);
				if(!Value.length())
					break;
			}
		}
	}*/
//	else
//	{
		for(i=0; i<(int)(m_vApplications.size()); i++)
		{
			NvApplication *Application = (NvApplication *)m_vApplications[i];
			if((Application != NULL) && (Application->GetName() != NV_APPLICATION_DEFAULT_NAME))
				AuthorizedApplications[Application]=1;
		}
//	}
	for(i=0; i<(int)(m_vApplications.size()); i++)
	{
		NvApplication *Application = (NvApplication *)(m_vApplications[i]);
		if(AuthorizedApplications.find(Application) != AuthorizedApplications.end())
			Application->Start(false, false);
		else if(Application->GetName() != NV_APPLICATION_DEFAULT_NAME)
			Application->m_bEnable = false;
	}
	NV_UNLOCKLOCK(&ProtectStartStop);
}


void NvApplicationList::Stop()
{
	int i;
	NV_LOCKOBJECT(this);
	vector<NvApplication*> ApplicationsToStop;
	for(i=0; i<(int)(m_vApplications.size()); i++)
	{
		NvApplication *Application = (NvApplication *)m_vApplications[i];
		if(Application != NULL)
			ApplicationsToStop.push_back(Application);
	}
	NV_UNLOCKOBJECT(this);
	NV_LOCKLOCK(&ProtectStartStop);
	for(i=0; i<(int)(ApplicationsToStop.size()); i++)
	{
		NvApplication *Application = ApplicationsToStop[i];
		Application->Stop();
	}
	NV_UNLOCKLOCK(&ProtectStartStop);
}


bool NvApplicationList::StopListeners()
{
	int i;
	for(i=0; i<(int)(m_vApplications.size()); i++)
	{
		NvApplication *Application = (NvApplication *)m_vApplications[i];
		try
		{
			Application->Listeners.StopListeners();
		}
		catch(...)
		{
		}
	}
	return true;
}


bool NvApplicationList::StartListeners()
{
	// Start listeners of all enabled applications
	// Not locked because called only from the nvs.cpp at initialisation time

	int i;
	for(i=0; i<(int)(m_vApplications.size()); i++)
	{
		NvApplication *Application = (NvApplication *)m_vApplications[i];
		try
		{
			if(Application->m_bEnable)
			{
				Application->Listeners.Load();
				Application->Listeners.StartListeners();
			}
		}
		catch(...)
		{
		}
	}

	return true;
}


NvApplication* NvApplicationList::InternalGetApplication(const char *Name, int *Index)
{
	string NameHere = NV_APPLICATION_DEFAULT_NAME;

	if(Name != NULL)
		NameHere = Name;
	
	NvApplication::CleanAppName(NameHere);

	if(!NameHere.length())
		NameHere = NV_APPLICATION_DEFAULT_NAME;

	MakeUpper(NameHere);
	NvApplication *FoundApplication = NULL;

	int i;
	for(i=0; i<(int)(m_vApplications.size()); i++)
	{
		NvApplication *Application = (NvApplication *)m_vApplications[i];
		if(Application == NULL)
			continue;
		if(NameHere == Application->GetName())
		{
			FoundApplication = Application;
			if(Index != NULL)
				*Index = i;
			break;
		}
	}

	return FoundApplication;
}


NvApplication* NvApplicationList::GetApplication(const char *Name, bool LockIt)
{
	if(Name == NULL)
		return NULL;
	NvApplication *ReturnedApplication = NULL;
	NV_LOCKOBJECT(this);
	ReturnedApplication = InternalGetApplication(Name);

	// Initialize the application if not yet done
	if(ReturnedApplication != NULL)
		if(!ReturnedApplication->Init())
			ReturnedApplication = NULL;

	if(LockIt && (ReturnedApplication != NULL))
		ReturnedApplication->IncInuse();
	NV_UNLOCKOBJECT(this);
	return ReturnedApplication;
}


bool NvApplicationList::EnableApplication(const char *Name, bool FromLoad)
{
	if(Name == NULL)
		return false;
	NvApplication *ReturnedApplication = NULL;
	NV_LOCKOBJECT(this);
	ReturnedApplication = InternalGetApplication(Name);

	if(ReturnedApplication == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	if(ReturnedApplication->bInInstall)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	bool RetVal = ReturnedApplication->Enable();
	if(!FromLoad)
		if(RetVal)
			Save();

	NV_UNLOCKOBJECT(this);
	return RetVal;
}


bool NvApplicationList::StartApplication(const char *Name)
{
	if(Name == NULL)
		return false;
	NvApplication *ReturnedApplication = NULL;
	NV_LOCKOBJECT(this);
	ReturnedApplication = InternalGetApplication(Name);

	if(ReturnedApplication == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	if(ReturnedApplication->bInInstall)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}
	// Get the list of authorized applications
/*	string Value = "";
	if(GetMain()->Licenses.CheckKey("SYSTEM",NVS_SERVICE_ID,"APPLICATIONS", Value, false))
	{
		bool bFound = false;
		// The license is application limited
		// Check if time limited
		if(!GetMain()->Licenses.CheckKey("SYSTEM",NVS_SERVICE_ID,"APPLICATIONS", Value))
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		// We are not time limited so we can run the authorized application
		string AppName = "";
		string::size_type index;
		while(1)
		{
			index = Value.find(';');
			if(index != string::npos)
			{
				AppName = Value.substr(0, index);
				Value = Value.substr(index+1);
			}
			else
			{
				AppName = Value;
				Value = "";
			}
			NvApplication::CleanAppName(AppName);
			NvApplication *Application = InternalGetApplication(AppName.c_str());
			if(Application == ReturnedApplication)
			{
				bFound = true;
				break;
			}
			CleanString(Value);
			if(!Value.length())
				break;
		}
		if(!bFound)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}
*/
	NV_UNLOCKOBJECT(this);
	NV_LOCKLOCK(&ProtectStartStop);
	bool RetVal = ReturnedApplication->Start(true);
	NV_UNLOCKLOCK(&ProtectStartStop);
	NV_LOCKOBJECT(this);
	if(RetVal)
		Save();

	NV_UNLOCKOBJECT(this);
	return RetVal;
}



bool NvApplicationList::StopApplication(const char *Name, const char *SessionId)
{
	if(Name == NULL)
		return false;
	NvApplication *ReturnedApplication = NULL;
	NV_LOCKOBJECT(this);
	ReturnedApplication = InternalGetApplication(Name);

	if(ReturnedApplication == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}
	string AppName = ReturnedApplication->GetName();
	MakeUpper(AppName);
	if(AppName == NV_APPLICATION_DEFAULT_NAME)
	{
		// Cannot disable default application
		NV_UNLOCKOBJECT(this);
		return false;
	}

	NV_UNLOCKOBJECT(this);
	NV_LOCKLOCK(&ProtectStartStop);
	bool RetVal = ReturnedApplication->Stop(SessionId, true);
	NV_UNLOCKLOCK(&ProtectStartStop);
	NV_LOCKOBJECT(this);
	if(RetVal)
		Save();
	NV_UNLOCKOBJECT(this);
	return RetVal;
}

bool NvApplicationList::CreateApplication(const char *Name, const char *Desc, int DefaultTimeOut, const char *SecuApp, bool ScheduledTransac, bool AllowWebXml, bool FromLoad, bool AllowNotOwnedSession, bool CreateLicense, int AuthMode, bool NirvaIsSsoUser)
{
	if(Name == NULL)
		return false;

	string NameHere = Name;
	NvApplication::CleanAppName(NameHere);

	if(!NameHere.length())
		return false;

	MakeUpper(NameHere);
	
	NV_LOCKOBJECT(this);
	
	// Test if the application exists
	NvApplication *Application = InternalGetApplication(NameHere.c_str());
	bool bNewApplication = true;
	if(Application != NULL)
	{
		if(Application->bInInstall)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(Application->IsEnable())
		{
			NV_UNLOCKOBJECT(this);
			return false;	// The application is actually running, we cannot change its parameters
		}

		bNewApplication = false;
	}
	else
	{
		Application = NVNEW NvApplication;
		if(Application == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	bool RetVal = Application->SetParameters(NameHere.c_str(), Desc, DefaultTimeOut, SecuApp, ScheduledTransac, AllowWebXml, AllowNotOwnedSession, AuthMode, NirvaIsSsoUser);
	if(!FromLoad)
	{
		if(!RetVal)
		{
			if(bNewApplication)
				delete Application;
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}

	if(CreateLicense)
		Application->CreateLicenseFile();
	if(bNewApplication)
	{
		m_vApplications.push_back((void *)Application);
		// Load logs
		Application->Logs.m_sAppName = Application->GetName();
		Application->Logs.Owner = (void *)Application;
		Application->Logs.Load();
	}

	if(!FromLoad)
	{
		if(!Save())
		{
			if(bNewApplication)
			{
				m_vApplications.pop_back();
				delete Application;
			}
			NV_UNLOCKOBJECT(this);
			return false;
		}
	}
	
	NV_UNLOCKOBJECT(this);
	return RetVal;
	
}


void NvApplicationList::FreeApplication(const char *Name)
{
	NvApplication *FoundApplication = NULL;
	NV_LOCKOBJECT(this);
	FoundApplication = InternalGetApplication(Name);
	if(FoundApplication != NULL)
		FoundApplication->DecInUse();
	NV_UNLOCKOBJECT(this);
}


bool NvApplicationList::RemoveApplication(const char *Name)
{
	NvApplication *FoundApplication = NULL;
	NV_LOCKOBJECT(this);
	int Index;
	FoundApplication = InternalGetApplication(Name, &Index);
	if(FoundApplication != NULL)
	{
		if(FoundApplication->bInInstall)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(FoundApplication->IsEnable() || FoundApplication->GetInUse())
		{
			// Cannot remove a running application
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(FoundApplication->GetName() != NV_APPLICATION_DEFAULT_NAME)
		{
			// We cannot remove the default application
			m_vApplications.erase(m_vApplications.begin()+Index);
			if(!Save())
			{
				m_vApplications.push_back((void *)FoundApplication);
				NV_UNLOCKOBJECT(this);
				return false;
			}
			delete FoundApplication;
		}
	}
	NV_UNLOCKOBJECT(this);
	return true;
}


bool NvApplicationList::ApplicationExist(const char *Name)
{
	NvApplication *FoundApplication = NULL;
	NV_LOCKOBJECT(this);
	FoundApplication = InternalGetApplication(Name);
	NV_UNLOCKOBJECT(this);
	if(FoundApplication == NULL)
		return false;

	return true;
}


bool NvApplicationList::IsApplicationEnable(const char *Name)
{
	NvApplication *FoundApplication = NULL;
	bool RetVal = false;
	NV_LOCKOBJECT(this);
	FoundApplication = InternalGetApplication(Name);
	if(FoundApplication != NULL)
		if(FoundApplication->IsEnable())
			RetVal = true;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}




bool NvApplicationList::GetSessions(const char *Name, NvObjStringList *ConnectedSessions)
{
	ConnectedSessions->RemoveAll();
	NvApplication *FoundApplication = NULL;
	bool RetVal = false;
	NV_LOCKOBJECT(this);
	FoundApplication = InternalGetApplication(Name);
	if(FoundApplication != NULL)
		if(FoundApplication->GetSessions(ConnectedSessions))
			RetVal = true;
	NV_UNLOCKOBJECT(this);

	return RetVal;
}



bool NvApplicationList::GetInfo(const char *Name, NvObjTable *ApplicationInfo)
{
	ApplicationInfo->Clear();

	bool AllApplications = true;

	NV_LOCKOBJECT(this);

	NvApplication *FoundApplication = NULL;
	if((Name != NULL) && strlen(Name))
	{
		// There is a request on a specific service
		FoundApplication = InternalGetApplication(Name);
		if(FoundApplication == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}

		AllApplications = false;
	}


	ApplicationInfo->InsertColumn("NAME");
	ApplicationInfo->InsertColumn("DESCRIPTION");
	ApplicationInfo->InsertColumn("TIMEOUT");
	ApplicationInfo->InsertColumn("SECUAPP");
	ApplicationInfo->InsertColumn("USE_SCHEDULED_TRANSAC");
	ApplicationInfo->InsertColumn("STATUS");
	ApplicationInfo->InsertColumn("NUM_SESSIONS");
	ApplicationInfo->InsertColumn("START_PAGE");
	ApplicationInfo->InsertColumn("ALLOW_WEB_XML");
	ApplicationInfo->InsertColumn("ALLOW_NOT_OWNED_SESSIONS");
	ApplicationInfo->InsertColumn("AUTH_MODE");
	ApplicationInfo->InsertColumn("NIRVA_IS_SSO_USER");

	if(AllApplications)
	{
		int i, j;
		for(i=0, j=0; i<(int)(m_vApplications.size()); i++)
		{
			NvApplication *Application = (NvApplication *)m_vApplications[i];
			if(Application == NULL)
				continue;
			ApplicationInfo->InsertRow();
			ApplicationInfo->InsertCellLine(j, 0, Application->GetName());
			ApplicationInfo->InsertCellLine(j, 1, Application->GetDesc());
			char Buffer[20];
			sprintf(Buffer, "%d", Application->GetDefaultTimeOut());
			ApplicationInfo->InsertCellLine(j, 2, Buffer);
			ApplicationInfo->InsertCellLine(j, 3, Application->GetSecuApp());
			if(Application->GetUseScheduledTransac())
				ApplicationInfo->InsertCellLine(j, 4, "YES");
			else
				ApplicationInfo->InsertCellLine(j, 4, "NO");
			if(Application->IsEnable())
				ApplicationInfo->InsertCellLine(j, 5, "RUNNING");
			else
				ApplicationInfo->InsertCellLine(j, 5, "STOPPED");
			sprintf(Buffer, "%d", Application->Sessions.GetSize());
			ApplicationInfo->InsertCellLine(j, 6, Buffer);
			ApplicationInfo->InsertCellLine(j, 7, Application->GetStartPage());
			if(Application->GetAllowWebXml())
				ApplicationInfo->InsertCellLine(j, 8, "YES");
			else
				ApplicationInfo->InsertCellLine(j, 8, "NO");
			if(Application->GetAllowNotOwnedSession())
				ApplicationInfo->InsertCellLine(j, 9, "YES");
			else
				ApplicationInfo->InsertCellLine(j, 9, "NO");
			int AuthMode = Application->GetAuthMode();
			strcpy(Buffer, "NIRVA");
			if(AuthMode == NV_AUTHENTICATION_SSO)
				strcpy(Buffer, "SSO");
			if(AuthMode == NV_AUTHENTICATION_NIRVA_SSO)
				strcpy(Buffer, "NIRVA_SSO");
			ApplicationInfo->InsertCellLine(j, 10, Buffer);
			if(Application->GetNirvaIsSsoUser())
				ApplicationInfo->InsertCellLine(j, 11, "YES");
			else
				ApplicationInfo->InsertCellLine(j, 11, "NO");
			j++;
		}
	}
	else
	{
		ApplicationInfo->InsertRow();
		ApplicationInfo->InsertCellLine(0, 0, FoundApplication->GetName());
		ApplicationInfo->InsertCellLine(0, 1, FoundApplication->GetDesc());
		char Buffer[20];
		sprintf(Buffer, "%d", FoundApplication->GetDefaultTimeOut());
		ApplicationInfo->InsertCellLine(0, 2, Buffer);
			ApplicationInfo->InsertCellLine(0, 3, FoundApplication->GetSecuApp());
		if(FoundApplication->GetUseScheduledTransac())
			ApplicationInfo->InsertCellLine(0, 4, "YES");
		else
			ApplicationInfo->InsertCellLine(0, 4, "NO");
		if(FoundApplication->IsEnable())
			ApplicationInfo->InsertCellLine(0, 5, "RUNNING");
		else
			ApplicationInfo->InsertCellLine(0, 5, "STOPPED");
		sprintf(Buffer, "%d", FoundApplication->Sessions.GetSize());
		ApplicationInfo->InsertCellLine(0, 6, Buffer);
		ApplicationInfo->InsertCellLine(0, 7, FoundApplication->GetStartPage());
		if(FoundApplication->GetAllowWebXml())
			ApplicationInfo->InsertCellLine(0, 8, "YES");
		else
			ApplicationInfo->InsertCellLine(0, 8, "NO");
		if(FoundApplication->GetAllowNotOwnedSession())
			ApplicationInfo->InsertCellLine(0, 9, "YES");
		else
			ApplicationInfo->InsertCellLine(0, 9, "NO");
		int AuthMode = FoundApplication->GetAuthMode();
		strcpy(Buffer, "NIRVA");
		if(AuthMode == NV_AUTHENTICATION_SSO)
			strcpy(Buffer, "SSO");
		if(AuthMode == NV_AUTHENTICATION_NIRVA_SSO)
			strcpy(Buffer, "NIRVA_SSO");
		ApplicationInfo->InsertCellLine(0, 10, Buffer);
		if(FoundApplication->GetNirvaIsSsoUser())
			ApplicationInfo->InsertCellLine(0, 11, "YES");
		else
			ApplicationInfo->InsertCellLine(0, 11, "NO");
	}
	NV_UNLOCKOBJECT(this);

	return true;
}



bool NvApplicationList::GetInfoEx(const char *Name, NvObjIndStringList *ApplicationInfo)
{
	if(Name == NULL)
		return false;

	if(!strlen(Name))
		return false;

	NV_LOCKOBJECT(this);

	NvApplication *FoundApplication = NULL;
	
	// Try to get the service
	FoundApplication = InternalGetApplication(Name);
	if(FoundApplication == NULL)
	{
		NV_UNLOCKOBJECT(this);
		return false;
	}

	FoundApplication->GetInfoEx(ApplicationInfo);

	NV_UNLOCKOBJECT(this);

	return true;
}





bool NvApplicationList::Install(const char *Name, const char *PackageFile, string& Info)
{
	if(Name == NULL)
		return false;

	bool Found = true;
	NvApplication *Application = NULL;
	NV_LOCKOBJECT(this);
	Application = InternalGetApplication(Name);
	string NameHere = Name;
	NvApplication::CleanAppName(NameHere);
	if(Application == NULL)
	{
		// The application has not been found so create a temporary service
		Application = NVNEW NvApplication;
		if(Application == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(!NameHere.length())
		{
			delete Application;
			NV_UNLOCKOBJECT(this);
			return false;
		}
		Application->m_sName = NameHere;
		Found = false;
	}
	else
	{
		if(Application->bInInstall)
		{
			Info = "The application is ever in install mode";
			NV_UNLOCKOBJECT(this);
			return false;
		}
		if(Application->IsEnable())
		{
			Info = "The application is running";
			NV_UNLOCKOBJECT(this);
			return false;	// The application is actually running, we cannot install it
		}
	}

	// Empty the xslt cache
	NvXml::EmptyCache();

	Application->bInInstall = true;
	NV_UNLOCKOBJECT(this);

	bool RetVal = true;
	
	try
	{
		RetVal = Application->Install(PackageFile, Info);
	}
	catch(...)
	{
		RetVal = false;
		Info = "Exception in application installation";
	}

	if(!Found)
		delete Application;

	if(Found)
	{
		NV_LOCKOBJECT(this);
		Application->bInInstall = false;
		NV_UNLOCKOBJECT(this);
	}

	if(RetVal && (!Found))
		RetVal = CreateApplication(NameHere.c_str(), "");

	return RetVal;
}


bool NvApplicationList::Package(const char *Name, const char *DescriptionFile, const char *PackageFile, string& Info)
{
	if(Name == NULL)
		return false;

	bool Found = true;
	NvApplication *Application = NULL;
	NV_LOCKOBJECT(this);
	Application = InternalGetApplication(Name);
	if(Application == NULL)
	{
		// The service has not been found so create a temporary service
		Application = NVNEW NvApplication;
		if(Application == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		string NameHere = Name;
		CleanString(NameHere);
		MakeUpper(NameHere);
		if(!NameHere.length())
		{
			delete Application;
			NV_UNLOCKOBJECT(this);
			return false;
		}
		Application->m_sName = NameHere;
		Found = false;
	}

	if(Application->bInInstall)
	{
		if(!Found)
			delete Application;
		NV_UNLOCKOBJECT(this);
		return false;
	}
	bool RetVal = Application->Package(DescriptionFile, PackageFile, Info);

	if(!Found)
		delete Application;

	NV_UNLOCKOBJECT(this);
	return RetVal;
}



/////////////////////////////////////////////////////////////////////////////
// NvApplication

NvApplication::NvApplication()
{
	Initialized = false;
	InUse = 0;
	m_bEnable = false;
	m_bStarted = false;
	m_sName = NV_APPLICATION_DEFAULT_NAME;
	m_sDesc = "";
	m_iDefaultTimeOut = 300;
	m_iAuthMode = NV_AUTHENTICATION_NIRVA;
	m_bNirvaIsSsoUser = false;
	m_sWorkDir = ""; 
	m_sRegistryDir = ""; 
	m_sDocDir = ""; 
	m_sFileDir = ""; 
	m_sLogDir = ""; 
	m_sWrootDir = ""; 
	m_sProcDir = ""; 
	m_sStartPage = "default.htm"; 
	AppRegistry.SetName("parameters");
	bInInstall = false;
	m_sSecurityApplication = "";
	bUseScheduledTransac = true;
	bAllowWebXml = true;
	bAllowNotOwnedSession = true;
	Containers.m_bUseLocking = true;
	Listeners.Application = this;
	Redirections.Application = this;
	m_bUseFramework = false;
}

NvApplication::~NvApplication()
{
	Cleanup();
	
	// Clear service information container
	NV_LOCKLOCK(&ProtectInUse);
	m_Info.Clear();
	NV_UNLOCKLOCK(&ProtectInUse);
}


bool NvApplication::Init()
{
	NV_LOCKOBJECT(this);
	if(Initialized)
	{
		NV_UNLOCKOBJECT(this);
		return true;
	}

	NV_LOCKLOCK(&ProtectInUse);
	if(!CreateApplicationDirectories())
	{
		NV_UNLOCKLOCK(&ProtectInUse);
		NV_UNLOCKOBJECT(this);
		return false;
	}

	NV_UNLOCKLOCK(&ProtectInUse);

	// We must load now the description file if it exists
	LoadInfo();

	// We set the application registry directory
	AppRegistry.SetRootDirectory(m_sRegistryDir.c_str());

	// Load test 	
	TestSets.m_sDirPath = m_sFileDir;
	TestSets.m_sDirPath += "Test";
	AddDirSuffix(TestSets.m_sDirPath);
	TestSets.m_sDirPath += "Sets";
	TestSets.Load();

	// Load transactions for this application
	Transactions.m_sAppName = GetName();
	// PM modif, the transaction list is only loaded on request
	//Transactions.Load();

	// Load tasks for this application
	Tasks.m_sAppName = GetName();
	// PM modif, the task list is only loaded on request
	//Tasks.Load();

	// Load security
	Security.m_Application = this;
	Security.Load();

	//Load redirections;
	Redirections.Load();

	NidgetFramework.m_sAppName = GetName();
	NvCommand CommandToExecute;	// Used only for errors
	CommandToExecute.m_sLanguage = "ENGLISH";
	if(GetUseFramework())
	{
		NidgetFramework.Skeleton();
		NidgetFramework.CheckPublish(&CommandToExecute);
	}
	// Do init code here
	Initialized = true;

	NV_UNLOCKOBJECT(this);

	return true;
}

void NvApplication::Cleanup()
{
	NV_LOCKOBJECT(this);
	if(!Initialized)
	{
		NV_UNLOCKOBJECT(this);
		return;
	}

	// Do cleanup code here

	AppRegistry.Clear();		// Removes the application registry containers from memory
	SemaList.Clear();			// Clears the semaphore list at the end because it can be used
								// by some other objects
	Containers.Clear();			// Clear application containers
	LockList.Clear();			// Clears the lock list at the end because it can be used
								// by some other objects
	NidgetFramework.Clear();	// Clear the framework

	Sessions.RemoveAll();

	Initialized = false;
	NV_UNLOCKOBJECT(this);
}


// Called from inside so it's not locked
bool NvApplication::CreateApplicationDirectories()
{
	string NameHere = m_sName;
	CleanAppName(NameHere);
	MakeUpper(NameHere);	// The previous function returns lower case
	if(!NameHere.length())
		return false;
	
	// Create app dir for nirva home
	string AppDirName = GetMain()->GetNirvaDir();
	AppDirName += "Applications";
	if(!CreateDir(AppDirName.c_str()))	// In fact, the directory should exist
		return false;
	AddDirSuffix(AppDirName);
	AppDirName += NameHere;
	if(!CreateDir(AppDirName.c_str()))
		return false;
	AddDirSuffix(AppDirName);

	// Create app dir for log home
	string AppHomeLogDirName = GetMain()->GetLogHomeDir();
	AppHomeLogDirName += "Applications";
	if(!CreateDir(AppHomeLogDirName.c_str()))	// In fact, the directory should exist
		return false;
	AddDirSuffix(AppHomeLogDirName);
	AppHomeLogDirName += NameHere;
	if(!CreateDir(AppHomeLogDirName.c_str()))
		return false;
	AddDirSuffix(AppHomeLogDirName);

	// Create app dir for registry home
	string AppHomeRegDirName = GetMain()->GetRegistryHomeDir();
	AppHomeRegDirName += "Applications";
	if(!CreateDir(AppHomeRegDirName.c_str()))	// In fact, the directory should exist
		return false;
	AddDirSuffix(AppHomeRegDirName);
	AppHomeRegDirName += NameHere;
	if(!CreateDir(AppHomeRegDirName.c_str()))
		return false;
	AddDirSuffix(AppHomeRegDirName);

	// Create app dir for work home
	string AppHomeWorkDirName = GetMain()->GetAppWorkHomeDir();
	AppHomeWorkDirName += "Applications";
	if(!CreateDir(AppHomeWorkDirName.c_str()))	// In fact, the directory should exist
		return false;
	AddDirSuffix(AppHomeWorkDirName);
	AppHomeWorkDirName += NameHere;
	if(!CreateDir(AppHomeWorkDirName.c_str()))
		return false;
	AddDirSuffix(AppHomeWorkDirName);
	
	// Create temp dir (work dir)
	m_sWorkDir = AppHomeWorkDirName;
	m_sWorkDir += "Work";
	if(!CreateDir(m_sWorkDir.c_str()))
	{
		m_sWorkDir = "";
		return false;
	}
	AddDirSuffix(m_sWorkDir);

	// Create registry dir
	m_sRegistryDir = AppHomeRegDirName;
	m_sRegistryDir += "Registry";
	if(!CreateDir(m_sRegistryDir.c_str()))
	{
		m_sRegistryDir = "";
		return false;
	}
	AddDirSuffix(m_sRegistryDir);

	// Create doc dir
	m_sDocDir = AppDirName;
	m_sDocDir += "Docs";
	if(!CreateDir(m_sDocDir.c_str()))
	{
		m_sDocDir = "";
		return false;
	}
	else
	{
		// Also create the Html subdirectory
		string HtmlDir = m_sDocDir;
		AddDirSuffix(HtmlDir);
		HtmlDir += "Html";
		if(!CreateDir(HtmlDir.c_str()))
			return false;
	}
	AddDirSuffix(m_sDocDir);

	// Create bin dir
	string sBinDir = AppDirName;
	sBinDir += "Bin";
	CreateDir(sBinDir.c_str());

	// Create file dir
	m_sFileDir = AppDirName;
	m_sFileDir += "Files";
	if(!CreateDir(m_sFileDir.c_str()))
	{
		m_sFileDir = "";
		return false;
	}
	else
	{
		// Also create the Config subdirectory
		string ConfigDir = m_sFileDir;
		AddDirSuffix(ConfigDir);
		ConfigDir += "Config";
		if(!CreateDir(ConfigDir.c_str()))
			return false;

		// And the Test and Test/Sets subdiretories
		string TestDir = m_sFileDir;
		AddDirSuffix(TestDir);
		TestDir += "Test";
		CreateDir(TestDir.c_str());
		AddDirSuffix(TestDir);
		TestDir += "Sets";
		CreateDir(TestDir.c_str());
	}
	AddDirSuffix(m_sFileDir);

	// Create log dir
	m_sLogDir = AppHomeLogDirName;
	m_sLogDir += "Logs";
	if(!CreateDir(m_sLogDir.c_str()))
	{
		m_sLogDir = "";
		return false;
	}
	AddDirSuffix(m_sLogDir);

	// Create wroot dir
	m_sWrootDir = AppDirName;
	m_sWrootDir += "Wroot";
	if(!CreateDir(m_sWrootDir.c_str()))
	{
		m_sWrootDir = "";
		return false;
	}
	else
	{
		// Also create the Config subdirectory
		string ConfigDir = m_sWrootDir;
		AddDirSuffix(ConfigDir);
		ConfigDir += "Config";
		if(!CreateDir(ConfigDir.c_str()))
			return false;
	}
	AddDirSuffix(m_sWrootDir);

	// Create proc dir
	m_sProcDir = AppDirName;
	m_sProcDir += "Procs";
	if(!CreateDir(m_sProcDir.c_str()))
	{
		m_sProcDir = "";
		return false;
	}
	else
	{
		// Also create the Config subdirectory
		string ConfigDir = m_sProcDir;
		AddDirSuffix(ConfigDir);
		ConfigDir += "Config";
		if(!CreateDir(ConfigDir.c_str()))
			return false;
	}
	AddDirSuffix(m_sProcDir);

	string ApplicationNameLower = NameHere;
	MakeLower(ApplicationNameLower);

	NvObjFile SourceFile;
	SourceFile.SetPersistTime();
	NvObjFile DestinationFile;
	DestinationFile.SetPersistTime();

	string FileDir = GetMain()->GetNirvaDir();
	FileDir += "Files";
	AddDirSuffix(FileDir);
	FileDir += "Application";
	AddDirSuffix(FileDir);
	FileDir += "New";
	AddDirSuffix(FileDir);
	

	// Description file
	string FileName = FileDir;
	FileName += "application.dsc";
	SourceFile.SetFileName(FileName);

	FileName = m_sFileDir;
	FileName += "application.dsc";
	DestinationFile.SetFileName(FileName);

	// .dsc file
	if(!DestinationFile.Exist())
	{
		// Check if there is no old application name description file
		string oldDescFileName = m_sFileDir;
		oldDescFileName += ApplicationNameLower;
		oldDescFileName += ".dsc";
		if(!FileExist(oldDescFileName.c_str()))
		{
			if(!DestinationFile.CopyFrom(&SourceFile))
				return false;
			if(!SetApplicationFile(&DestinationFile))
				return false;
		}
	}

	// Default doc file
	FileName = FileDir;
	FileName += "doc.htm";
	SourceFile.SetFileName(FileName);

	FileName = m_sDocDir;
	FileName += "Html";
	AddDirSuffix(FileName);
	FileName += "index.html";
	DestinationFile.SetFileName(FileName);

	// default.htm file
	if(!DestinationFile.Exist())
	{
		if(!DestinationFile.CopyFrom(&SourceFile))
			return false;
		if(!SetApplicationFile(&DestinationFile))
			return false;
	}

	// Config procedure file
	FileName = FileDir;
	FileName += "config.nvp";
	SourceFile.SetFileName(FileName);

	FileName = m_sProcDir;
	FileName += "Config";
	AddDirSuffix(FileName);
	FileName += "config.nvp";
	DestinationFile.SetFileName(FileName);

	// .nvp file
	if(!DestinationFile.Exist())
	{
		if(!DestinationFile.CopyFrom(&SourceFile))
			return false;
		if(!SetApplicationFile(&DestinationFile))
			return false;
	}

	// Config xsl file
	FileName = FileDir;
	FileName += "config.xsl";
	SourceFile.SetFileName(FileName);

	FileName = m_sFileDir;
	FileName += "Config";
	AddDirSuffix(FileName);
	FileName += "config.xsl";
	DestinationFile.SetFileName(FileName);

	if(!DestinationFile.Exist())
	{
		if(!DestinationFile.CopyFrom(&SourceFile))
			return false;
		if(!SetApplicationFile(&DestinationFile))
			return false;
	}

	// package file
	FileName = FileDir;
	FileName += "package.lst";
	SourceFile.SetFileName(FileName);

	FileName = m_sFileDir;
	FileName += "package.lst";
	DestinationFile.SetFileName(FileName);

	if(!DestinationFile.Exist())
	{
		if(!DestinationFile.CopyFrom(&SourceFile))
			return false;
		if(!SetApplicationFile(&DestinationFile))
			return false;
	}

	// default.htm file
	FileName = FileDir;
	FileName += "default.htm";
	SourceFile.SetFileName(FileName);

	FileName = m_sWrootDir;
	FileName += "default.htm";
	DestinationFile.SetFileName(FileName);

	if(!DestinationFile.Exist())
	{
		// No error if no success
		if(DestinationFile.CopyFrom(&SourceFile))
			SetApplicationFile(&DestinationFile);
	}

	// robots.txt file
	FileName = FileDir;
	FileName += "robots.txt";
	SourceFile.SetFileName(FileName);

	FileName = m_sWrootDir;
	FileName += "robots.txt";
	DestinationFile.SetFileName(FileName);

	if(!DestinationFile.Exist())
	{
		// No error if no success
		if(DestinationFile.CopyFrom(&SourceFile))
			SetApplicationFile(&DestinationFile);
	}

	// init.nvp file
	FileName = FileDir;
	FileName += "init.nvp";
	SourceFile.SetFileName(FileName);

	FileName = m_sProcDir;
	FileName += "init.nvp";
	DestinationFile.SetFileName(FileName);

	if(!DestinationFile.Exist())
	{
		// No error if no success
		if(DestinationFile.CopyFrom(&SourceFile))
			SetApplicationFile(&DestinationFile);
	}

	// exit.nvp file
	FileName = FileDir;
	FileName += "exit.nvp";
	SourceFile.SetFileName(FileName);

	FileName = m_sProcDir;
	FileName += "exit.nvp";
	DestinationFile.SetFileName(FileName);

	if(!DestinationFile.Exist())
	{
		// No error if no success
		if(DestinationFile.CopyFrom(&SourceFile))
			SetApplicationFile(&DestinationFile);
	}

	// session_open.nvp file
	FileName = FileDir;
	FileName += "session_open.nvp";
	SourceFile.SetFileName(FileName);

	FileName = m_sProcDir;
	FileName += "session_open.nvp";
	DestinationFile.SetFileName(FileName);

	if(!DestinationFile.Exist())
	{
		// No error if no success
		if(DestinationFile.CopyFrom(&SourceFile))
			SetApplicationFile(&DestinationFile);
	}

	// session_close.nvp file
	FileName = FileDir;
	FileName += "session_close.nvp";
	SourceFile.SetFileName(FileName);

	FileName = m_sProcDir;
	FileName += "session_close.nvp";
	DestinationFile.SetFileName(FileName);

	if(!DestinationFile.Exist())
	{
		// No error if no success
		if(DestinationFile.CopyFrom(&SourceFile))
			SetApplicationFile(&DestinationFile);
	}


	return true;
}


// Called from inside so it's not locked
bool NvApplication::CreateLicenseFile()
{
	string NameHere = m_sName;
	CleanAppName(NameHere);
	MakeUpper(NameHere);	// The previous function returns lower case
	if(!NameHere.length())
		return false;
	
	// Create app dir for nirva home
	string AppDirName = GetMain()->GetNirvaDir();
	AppDirName += "Applications";
	AddDirSuffix(AppDirName);
	AppDirName += NameHere;
	AddDirSuffix(AppDirName);


	NvObjFile SourceFile;
	SourceFile.SetPersistTime();
	NvObjFile DestinationFile;
	DestinationFile.SetPersistTime();

	string FileDir = GetMain()->GetNirvaDir();
	FileDir += "Files";
	AddDirSuffix(FileDir);
	FileDir += "Application";
	AddDirSuffix(FileDir);
	FileDir += "New";
	AddDirSuffix(FileDir);
	
	
	// license file
	string FileName = FileDir;
	FileName += "license.txt";
	SourceFile.SetFileName(FileName);

	FileName = AppDirName;
	FileName += "license.txt";
	DestinationFile.SetFileName(FileName);

	NV_LOCKLOCK(&ProtectInUse);
	// license file
	if(!DestinationFile.Exist())
	{
		if(DestinationFile.CopyFrom(&SourceFile))
			if(!SetApplicationFile(&DestinationFile))
			{
				NV_UNLOCKLOCK(&ProtectInUse);
				return false;
			}
	}


	NV_UNLOCKLOCK(&ProtectInUse);
	return true;
}


// used for automatic source code generation
bool NvApplication::SetApplicationFile(NvObjFile *File)
{
	
	string AppWorkDir = m_sWorkDir;

	if(!File->Open("r+b"))
	{
		File->Remove();
		return false;
	}

	NvObjFile ProviFile;
	ProviFile.SetAutoFileName(AppWorkDir.c_str());
	if(!ProviFile.Open("w+b"))
	{
		File->Close();
		File->Remove();
		return false;
	}
	
	string Line;
	string::size_type index;

	string ApplicationNameUpper = m_sName;
	CleanAppName(ApplicationNameUpper);
	MakeUpper(ApplicationNameUpper);
	
	string ApplicationNameLower = m_sName;
	CleanAppName(ApplicationNameLower);
	MakeLower(ApplicationNameLower);


	// Create application identifier
	char Buffer[128];
	sprintf(Buffer, "%d%d%d", rand(), rand(), rand());
	string sIdentifier = Buffer;
	sIdentifier += m_sName;
	sIdentifier += m_sDesc;
	sIdentifier += m_sWorkDir;
	sIdentifier = MD5_string(sIdentifier);
	MakeUpper(sIdentifier);
	string sIdentifierForCheck = MD5_string(sIdentifier);
	MakeUpper(sIdentifierForCheck);

	while(File->ReadLine(Line))
	{

		while((index = Line.find("upperapplicationname")) != string::npos)
			Line.replace(index, 20, ApplicationNameUpper);
		while((index = Line.find("lowerapplicationname")) != string::npos)
			Line.replace(index, 20, ApplicationNameLower);
		while((index = Line.find("nirvaapplicationidcheck")) != string::npos)
			Line.replace(index, 23, sIdentifierForCheck);
		while((index = Line.find("nirvaapplicationid")) != string::npos)
			Line.replace(index, 18, sIdentifier);

#ifdef WIN32
		Line += "\r\n";
#else
		Line += "\n";
#endif
		
		if(ProviFile.Write(Line.c_str(), (int)(Line.length())) != (int)(Line.length()))
		{
			ProviFile.Close();
			File->Close();
			File->Remove();
			return false;
		}
	}

	ProviFile.Close();
	File->Close();
	
	if(!File->CopyFrom(&ProviFile))
	{
		File->Remove();
		return false;
	}
		
	return true;
}


// Load application information from dsc file
bool NvApplication::LoadInfo()
{
	// The description file should be in the File directory
	string ApplicationNameLower = GetName();
	MakeLower(ApplicationNameLower);
	string DescFileName = m_sFileDir;
	DescFileName += "application.dsc";

	NV_LOCKLOCK(&ProtectInUse);
	// If the application.dsc doesn't exist, use the real application name (compatibility with old versions)
	if(!GetNoCaseRealFileName(DescFileName) || !FileExist(DescFileName.c_str()))
	{
		DescFileName = m_sFileDir;
		DescFileName += ApplicationNameLower;
		DescFileName += ".dsc";
	}
	
	if(GetNoCaseRealFileName(DescFileName))
	{
		if(FileExist(DescFileName.c_str()))
		{
			if(!m_Info.Load(DescFileName.c_str()))
			{
				NV_UNLOCKLOCK(&ProtectInUse);
				return false;
			}
		}
	}

	if(!m_sDesc.length())
	{
		NvObjIndStringList *IndStringList = (NvObjIndStringList *)(m_Info.GetObject("INFO"));
		if(IndStringList != NULL)
		{
			if(IndStringList->KeyExist("DESCRIPTION"))
				m_sDesc = IndStringList->GetValue("DESCRIPTION");
			else if(IndStringList->KeyExist("Description"))
				m_sDesc = IndStringList->GetValue("Description");
			else if(IndStringList->KeyExist("description"))
				m_sDesc = IndStringList->GetValue("description");
		}
	}
	NvObjIndStringList *IndStringList = (NvObjIndStringList *)(m_Info.GetObject("SETTINGS"));
	if(IndStringList != NULL)
	{
		string StartPage = IndStringList->GetValue("StartPage");
		if(StartPage.length())
			m_sStartPage = StartPage;
		string UseFramework = IndStringList->GetValue("Framework");
		CleanString(UseFramework);
		MakeUpper(UseFramework);
		if(UseFramework == "YES")
			m_bUseFramework = true;
	}
	NV_UNLOCKLOCK(&ProtectInUse);
	return true;
}




bool NvApplication::GetInfoEx(NvObjIndStringList *ApplicationInfo)
{
	if(ApplicationInfo == NULL)
		return false;

	ApplicationInfo->RemoveAll();

	// We must load now the description file if it exists
	LoadInfo();

	NV_LOCKLOCK(&ProtectInUse);
	NvObjIndStringList *IndStringList = (NvObjIndStringList *)(m_Info.GetObject("INFO"));
	if(IndStringList != NULL)
		ApplicationInfo->CopyFrom(IndStringList);
	NV_UNLOCKLOCK(&ProtectInUse);

	return true;
}

void NvApplication::IncInuse()
{
	NV_LOCKLOCK(&ProtectInUse);
	InUse++;
	NV_UNLOCKLOCK(&ProtectInUse);
}

void NvApplication::DecInUse()
{
	NV_LOCKLOCK(&ProtectInUse);
	if(InUse)
		InUse--;
	NV_UNLOCKLOCK(&ProtectInUse);
}

int NvApplication::GetInUse()
{
	int RetVal = 0;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = InUse;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}



bool NvApplication::Enable()
{
	if(!Init())
		return false;
	
	// If is enabled, return immediatly OK
	if(IsEnable())
		return true;

	NV_LOCKLOCK(&ProtectInUse);
	m_bEnable = true;
	NV_UNLOCKLOCK(&ProtectInUse);

	NV_LOCKLOCK(&ProtectInUse);
	if(!CreateApplicationDirectories())
	{
		NV_UNLOCKLOCK(&ProtectInUse);
		NV_UNLOCKOBJECT(this);
		return false;
	}

	NV_UNLOCKLOCK(&ProtectInUse);

	// We must load now the description file if it exists
	LoadInfo();

	// We also load again the test sets in cas they have changed
	TestSets.Load();

	return true;
}



// Run the init procedure
bool NvApplication::Start(bool EnableIt, bool WithListeners)
{
	// First enable the application
	if(EnableIt)
		if(!Enable())
			return false;

	if(!IsEnable())
		return false;
	
	bool RetVal = true;
	if(!m_bStarted)
	{
		// Load the listeners before starting the application
		Listeners.Load();

		// Call here the application init procedure
		// From a special administrator session
		// We do not need any login here
		NvSession Session;
		Session.Init();
		Session.SetAppName(GetName().c_str());
		Session.Application = this;
		Session.m_bAdministrator = true;

		NvCommand Command;
		Session.OriginalCommand = &Command;
		Command.m_pSession = &Session;
		Command.m_iCommandSource = NV_SOURCE_SERVER;
		Command.m_sService = "SYSTEM";
		Command.m_sClass = "MISC";
		Command.m_sCommand = "NOP";
		if(!Session.ExecuteProcedure(&Command, "init", false))
			RetVal = false;
		else
			m_bStarted = true;
		//Session.Application = NULL;
	}

	if(!RetVal)
	{
		// We disable the application if cannot be started
		NV_LOCKLOCK(&ProtectInUse);
		m_bEnable = false;
		NV_UNLOCKLOCK(&ProtectInUse);
		SLEEP(100);
		GetMain()->Sessions.Clear(GetName().c_str());
	}

	// Start the listeners for this application
	if(WithListeners)
		Listeners.StartListeners();

	return RetVal;
}




// Stop and eventually Disable the application and closes all sessions using it
// SessionId is the session id of the session launching the Stop application command
bool NvApplication::Stop(const char *SessionId, bool DisableIt)
{
	// If not enabled, return immediatly OK
	if(!IsEnable())
	{
		m_bStarted = false;
		return true;
	}

	// Return an error if an application tries to disable itself
	if(SessionId != NULL)
	{
		string SessionIdHere = SessionId;
		CleanString(SessionIdHere);
		MakeUpper(SessionIdHere);
		
		NV_LOCKOBJECT(this);
		if(Sessions.KeyExist(SessionIdHere))
		{
			// An application cannot disable itself
			NV_UNLOCKOBJECT(this);
			return false;
		}
		NV_UNLOCKOBJECT(this);
	}

	// Disable the application now if required
	if(DisableIt)
	{
		NV_LOCKLOCK(&ProtectInUse);
		m_bEnable = false;
		NV_UNLOCKLOCK(&ProtectInUse);
		SLEEP(100);
	}

	// Stop the listeners for this application
	try
	{
		Listeners.StopListeners();
	}
	catch(...)
	{
	}
	

	// Close here all the sessions connected to the application
	// This waits until all sessions for this application to be closed
	GetMain()->Sessions.Clear(GetName().c_str());

	// PM TODO
	// Call here the application exit procedure
	// Then call the stop procedures of app sessions
	// In reverse order


	if(m_bStarted)
	{
		// Call here the application exit procedure
		// From a special administrator session
		// We do not need any login here
		NvSession Session;
		Session.Init();
		Session.SetAppName(GetName().c_str());
		Session.Application = this;
		Session.m_bAdministrator = true;

		NvCommand Command;
		Session.OriginalCommand = &Command;
		Command.m_pSession = &Session;
		Command.m_iCommandSource = NV_SOURCE_SERVER;
		Command.m_sService = "SYSTEM";
		Command.m_sClass = "MISC";
		Command.m_sCommand = "NOP";
		Session.ExecuteProcedure(&Command, "exit", false);
		m_bStarted = false;
		//Session.Application = NULL;
	}

	NV_LOCKLOCK(&ProtectInUse);
	InUse = 0;
	NV_UNLOCKLOCK(&ProtectInUse);
	m_bStarted = false;
	
	NV_LOCKOBJECT(this);
	AppRegistry.Clear();		// Removes the application registry containers from memory
	SemaList.Clear();			// Clears the semaphore list at the end because it can be used
								// by some other objects
	Containers.Clear();			// Clear application containers
	Registries.Clear();			// Clear all user registries for this service
	LockList.Clear();			// Clears the lock list at the end because it can be used
								// by some other objects
	TestSets.Clear();			// Clear the test set list
	NidgetFramework.Clear();	// Clear the framework

	//PMARC 07/10
	Sessions.RemoveAll();

	// Free the dotnet AppDomain
	GetMain()->DotnetInterface.UnloadDomain(DOMAIN_TYPE_APPLICATION, GetName().c_str());

	NV_UNLOCKOBJECT(this);
	return true;
}


bool NvApplication::IsEnable()
{
	bool RetVal = true;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_bEnable;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvApplication::GetName()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sName;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvApplication::GetDesc()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sDesc;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvApplication::GetStartPage()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sStartPage;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


bool NvApplication::GetUseFramework()
{
	bool RetVal = false;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_bUseFramework;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}

int NvApplication::GetDefaultTimeOut()
{
	int RetVal = 300;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_iDefaultTimeOut;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}

int NvApplication::GetAuthMode()
{
	int RetVal = 300;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_iAuthMode;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}

bool NvApplication::GetNirvaIsSsoUser()
{
	bool RetVal = false;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_bNirvaIsSsoUser;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}

string NvApplication::GetSecuApp()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sSecurityApplication;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}

bool NvApplication::GetUseScheduledTransac()
{
	bool RetVal = true;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = bUseScheduledTransac;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}

bool NvApplication::GetAllowWebXml()
{
	bool RetVal = true;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = bAllowWebXml;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}

bool NvApplication::GetAllowNotOwnedSession()
{
	bool RetVal = true;
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = bAllowNotOwnedSession;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvApplication::GetWorkDir()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sWorkDir;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvApplication::GetRegistryDir()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sRegistryDir;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvApplication::GetDocDir()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sDocDir;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvApplication::GetFileDir()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sFileDir;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvApplication::GetLogDir()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sLogDir;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvApplication::GetWrootDir()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sWrootDir;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvApplication::GetProcDir()
{
	string RetVal = "";
	NV_LOCKLOCK(&ProtectInUse);
	RetVal = m_sProcDir;
	NV_UNLOCKLOCK(&ProtectInUse);

	return RetVal;
}


string NvApplication::GetAppNameForFiles()
{
	string AppNameHere = GetName();
	CleanAppName(AppNameHere);
	MakeLower(AppNameHere);

	return AppNameHere;
}



void NvApplication::CleanAppName(string& Name)
{
	MakeUpper(Name);
	CleanString(Name, true);
	string::size_type index;
	// remove backslashes
	while((index = Name.find('\\')) != string::npos)
		Name.erase(index, 1);
	// remove slashes
	while((index = Name.find('/')) != string::npos)
		Name.erase(index, 1);
	// remove double quotes
	while((index = Name.find('\"')) != string::npos)
		Name.erase(index, 1);
	// remove quotes
	while((index = Name.find('\'')) != string::npos)
		Name.erase(index, 1);
}




bool NvApplication::SetParameters(const char *Name, const char *Desc, int DefaultTimeOut, const char *SecuApp, bool ScheduledTransac, bool AllowWebXml, bool AllowNotOwnedSession, int AuthMode, bool NirvaIsSsoUser)
{
	if(Name == NULL)
		return false;

	string NameHere = Name;
	CleanString(NameHere);
	MakeUpper(NameHere);
	if(!NameHere.length())
		return false;
	
	NV_LOCKLOCK(&ProtectInUse);
	m_sName = NameHere;
	string DescHere = "";
	if(Desc != NULL)
		DescHere = Desc;
	CleanString(DescHere);
	m_sDesc = DescHere;
	m_iDefaultTimeOut = DefaultTimeOut;
	if(m_iDefaultTimeOut<0)
		m_iDefaultTimeOut = 3600;
	if(m_iDefaultTimeOut<5)
		m_iDefaultTimeOut = 5;
	m_iAuthMode = AuthMode;
	if(m_iAuthMode < 0)
		m_iAuthMode = NV_AUTHENTICATION_NIRVA;
	if(m_iAuthMode > NV_AUTHENTICATION_MAXVALUE)
		m_iAuthMode = NV_AUTHENTICATION_NIRVA;
	m_bNirvaIsSsoUser = NirvaIsSsoUser;

	bUseScheduledTransac = ScheduledTransac;
	bAllowWebXml = AllowWebXml;
	bAllowNotOwnedSession = AllowNotOwnedSession;

	if(SecuApp != NULL)
	{
		m_sSecurityApplication = SecuApp;
		CleanAppName(m_sSecurityApplication);
		string ActualAppName = m_sName;
		CleanAppName(ActualAppName);
		if(ActualAppName == m_sSecurityApplication)
			m_sSecurityApplication = "";
	}

	AppRegistry.SetName("parameters");

	if(!CreateApplicationDirectories())
	{
		NV_UNLOCKLOCK(&ProtectInUse);
		return false;
	}

	NV_UNLOCKLOCK(&ProtectInUse);

	// We must load now the description file if it exists
	LoadInfo();

	// We set the application root directory
	NV_LOCKLOCK(&ProtectInUse);
	AppRegistry.SetRootDirectory(m_sRegistryDir.c_str());
	NV_UNLOCKLOCK(&ProtectInUse);


	return true;
}




bool NvApplication::GetSessions(NvObjStringList *ConnectedSessions)
{
	ConnectedSessions->RemoveAll();
	NV_LOCKOBJECT(this);
	map<string,string>::iterator pos;
	for(pos = Sessions.m_mList.begin(); pos != Sessions.m_mList.end(); ++pos)
	{
		string SessionIdInList = pos->first;
		ConnectedSessions->Add(SessionIdInList);
	}
	NV_UNLOCKOBJECT(this);

	return true;
}




bool NvApplication::GetPermissions(NvObjTable *ApplicationPermissions)
{
	NvApplication *Application = this;
	string ActualSecuApp = GetSecuApp();
	if(ActualSecuApp.length())
	{
		if(ActualSecuApp == "(SYSTEM)")
		{
			return true;	// No system permissions here
		}

		// Other application security
		if(!GetMain()->Applications.IsApplicationEnable(ActualSecuApp.c_str()))
		{
			return false;	// appli not enabled
		}

		Application = GetMain()->Applications.GetApplication(ActualSecuApp.c_str());

		// The application doesn't exist
		if(Application == NULL)
		{
			return false;
		}
	}
	
	NV_LOCKLOCK(&ProtectInUse);
	NvObjIndStringList *IndStringList = (NvObjIndStringList *)(Application->m_Info.GetObject("PERMISSIONS"));
	if(IndStringList != NULL)
	{
		map<string,string>::iterator pos;
		for(pos = IndStringList->m_mList.begin(); pos != IndStringList->m_mList.end(); ++pos)
		{
			string KeyFromList = pos->first.c_str();
			CleanString(KeyFromList);
			MakeUpper(KeyFromList);
			if(!KeyFromList.length())
				continue;
			string ValueFromList = pos->second.c_str();
			CleanString(ValueFromList);
			if(!ValueFromList.length())
				continue;
			ApplicationPermissions->InsertRow();
			ApplicationPermissions->InsertCellLine(-1, 0, KeyFromList);
			ApplicationPermissions->InsertCellLine(-1, 1, ValueFromList);
		}
	}
	NV_UNLOCKLOCK(&ProtectInUse);
	
	if(ActualSecuApp.length())
		GetMain()->Applications.FreeApplication(ActualSecuApp.c_str());

	return true;
}


// return 1 if OK, 0 for general error, -1 if bad user, -2 if cannot get the security context, -3 if bad password, -4 if application not enabled, -5 if password must be changed, -6 if bad password format
int NvApplication::Login(const char *SessionId, const char *UserName, const char *UserPassword, const char *NewPassword, const char *NewPasswordConfirm, bool bWithNewPassword, bool bWithNewPasswordConfirm, NvObjIndStringList *UserPermissions, string& UserFullName, bool ControlPassword)
{
	UserFullName = "";
	//PM TODO
	// If UserName and UserPassword are both NULLS, this is and internal login from Nirva
	// So we'll authorize all permissions (internal super user)
	if(!Init())
		return 0;

	if(!IsEnable())
		return -4;

	

	// Do the real login here
	string SessionIdHere = SessionId;
	CleanString(SessionIdHere);
	MakeUpper(SessionIdHere);
	string UserNameHere = "";
	if(UserName != NULL)
		UserNameHere = UserName;
	CleanString(UserNameHere, true);
	MakeLower(UserNameHere);
	if(!UserNameHere.length())
		UserNameHere = NV_SESSION_DEFAULT_USER;

	if((UserName == NULL) && (UserPassword == NULL))
	{
		// internal super user
		NV_LOCKOBJECT(this);
		Sessions.SetValue(SessionIdHere, UserNameHere.c_str());
		NV_UNLOCKOBJECT(this);
		return 1;
	}


	string PasswordHere = "";
	string PasswordHereMD5 = "";
	if(UserPassword != NULL)
		PasswordHere = UserPassword;
	CleanString(PasswordHere, true);
	if(PasswordHere.length())
		 PasswordHereMD5 = MD5_string(PasswordHere);
	MakeUpper(PasswordHere);
	MakeUpper(PasswordHereMD5);
	if(UserNameHere == "nvadmin")
	{
		// NIRVA administrator
		// We must check the password from the main nvs object
		if(!GetMain()->GetAdminPassword().length())
		{
			// internal super user
			NV_LOCKOBJECT(this);
			Sessions.SetValue(SessionIdHere, UserNameHere.c_str());
			NV_UNLOCKOBJECT(this);
			return 1;
		}
		string AdminPassword = GetMain()->GetAdminPassword();
		if(ControlPassword && (AdminPassword != PasswordHereMD5))
		{
			// We can test also with original password (not MD5 coded)
			if(ControlPassword && (AdminPassword != PasswordHere))
			{
				// We can test also with original password (not MD5 coded)
				// internal super user
				return -3;	// bad password
			}
		}
		UserFullName = "Nirva administrator";
	}
	else
	{

		if(!IsLocalSecurity())
		{
			int RetVal = ExternalLogin(SessionId, UserName, UserPassword, NewPassword, NewPasswordConfirm, bWithNewPassword, bWithNewPasswordConfirm, UserPermissions, UserFullName, ControlPassword);
			return RetVal;
		}

		string SecurityService = GetServiceSecurity();
		if(SecurityService.length())
		{
			int RetVal = ServiceLogin(SessionId, UserName, UserPassword, NewPassword, NewPasswordConfirm, bWithNewPassword, bWithNewPasswordConfirm, UserPermissions, UserFullName, ControlPassword);
			return RetVal;
		}

		// We verify the user from the security
		NvSecurity *CurrentSecurity = &Security;
		NvApplication *Application = NULL;
		string ActualSecuApp = GetSecuApp();
		if(ActualSecuApp.length())
		{
			if(ActualSecuApp == "(SYSTEM)")
				CurrentSecurity = &(GetMain()->Security);
			else
			{
				// Other application security
				if(!GetMain()->Applications.IsApplicationEnable(ActualSecuApp.c_str()))
				{
					return -4;	// appli not enabled
				}

				Application = GetMain()->Applications.GetApplication(ActualSecuApp.c_str());

				// The application doesn't exist
				if(Application == NULL)
				{
					return -2;
				}
				CurrentSecurity = &(Application->Security);
			}
		}

		bool bNewPasswordProcResult = true;
		string ProcCheckPassword = CurrentSecurity->GetProcCheckPassword();
		if(bWithNewPassword && ProcCheckPassword.length())
		{
			string OldPasswordHere = "";
			if(UserPassword != NULL)
				OldPasswordHere = UserPassword;
			CleanString(OldPasswordHere);
			string NewPasswordHere = "";
			if(NewPassword != NULL)
				NewPasswordHere = NewPassword;
			CleanString(NewPasswordHere);

			// Call here the application check password procedure
			// From a special administrator session
			// We do not need any login here
			NvSession Session;
			Session.Init();
			Session.SetAppName(GetName().c_str());
			Session.Application = this;
			Session.m_bAdministrator = true;

			NvCommand Command;
			Session.OriginalCommand = &Command;
			Command.m_pSession = &Session;
			Command.m_iCommandSource = NV_SOURCE_SERVER;
			Command.m_sService = "SYSTEM";
			Command.m_sClass = "MISC";
			Command.m_sCommand = "NOP";
			Command.Parameters.SetValue("OLD_PASSWORD", OldPasswordHere);
			Command.Parameters.SetValue("NEW_PASSWORD", NewPasswordHere);
			Command.Parameters.SetValue("USER", UserNameHere);
			if(!Session.ExecuteProcedure(&Command, ProcCheckPassword, false))
				bNewPasswordProcResult = false;
			//Session.Application = NULL;
		}

		// Do the login at security level
		int SecuLoginRetVal = CurrentSecurity->Login(UserName, UserPassword, NewPassword, NewPasswordConfirm, bWithNewPassword, bWithNewPasswordConfirm, UserPermissions, UserFullName, ControlPassword, bNewPasswordProcResult);
		if(Application != NULL)
			GetMain()->Applications.FreeApplication(Application->GetName().c_str());
		int RetVal = 1;
		if(SecuLoginRetVal == -1)
			RetVal = -1;
		if(SecuLoginRetVal == -2)
			RetVal = -1;
		if(SecuLoginRetVal == -3)
			RetVal = -3;
		if(SecuLoginRetVal == -4)
			RetVal = -5;
		if(SecuLoginRetVal == -5)
			RetVal = -6;
		if(RetVal < 0)
			return RetVal;
	}

	NV_LOCKOBJECT(this);
	Sessions.SetValue(SessionIdHere, UserNameHere.c_str());
	NV_UNLOCKOBJECT(this);

	return 1;
}


// Do the login on another application on another server
// return 1 if OK, 0 for general error, -1 if bad user, -2 if cannot get the security context, -3 if bad password, -4 if application not enabled, -5 if password must be changed, -6 if bad new password syntax
int NvApplication::ExternalLogin(const char *SessionId, const char *UserName, const char *UserPassword, const char *NewPassword, const char *NewPasswordConfirm, bool bWithNewPassword, bool bWithNewPasswordConfirm, NvObjIndStringList *UserPermissions, string& UserFullName, bool ControlPassword)
{
	// Do the real login here
	string SessionIdHere = SessionId;
	CleanString(SessionIdHere);
	MakeUpper(SessionIdHere);
	string UserNameHere = "";
	if(UserName != NULL)
		UserNameHere = UserName;
	CleanString(UserNameHere, true);
	MakeLower(UserNameHere);
	if(!UserNameHere.length())
		UserNameHere = NV_SESSION_DEFAULT_USER;

	// Get the server adress port and app name and SSL flag
	// The secu app have the format APPNAME[server_address:server_port(SSL)]
	// or APPNAME[server_address:server_port] if not using SSL
	string ActualSecuApp = GetSecuApp();
	string::size_type index = ActualSecuApp.find('[');
	string ServerAdress = ActualSecuApp.substr(index+1);
	ActualSecuApp = ActualSecuApp.substr(0, index);
	CleanString(ActualSecuApp);
	index = ServerAdress.find(']');
	ServerAdress = ServerAdress.substr(0, index);
	CleanString(ServerAdress);
	bool bUseSsl = false;
	if(ServerAdress.length() >= 5)
	{
		string sUseSsl = ServerAdress.substr(ServerAdress.length()-5);
		MakeUpper(sUseSsl);
		if(sUseSsl == "(SSL)")
		{
			bUseSsl = true;
			ServerAdress = ServerAdress.substr(0, ServerAdress.length()-5);
		}
	}
	CleanString(ServerAdress);



	// Create a new temporary session just for login
	NvSession *SpecialSession = NVNEW NvSession;
	SpecialSession->Init();
	SpecialSession->m_bAdministrator = true;
	
	// We use request to connect to peer
	NvRequest SpecialRequest;
	SpecialRequest.Session = SpecialSession;
	SpecialRequest.b_UseSSL = bUseSsl;
	SpecialRequest.b_ClientUnicode = GetMain()->IsUnicode();

	SpecialRequest.m_sServer = ServerAdress;
	SpecialRequest.m_sAppli = ActualSecuApp;
	SpecialRequest.m_sOpen = "NV_SESSION_OPEN_NONE";
	SpecialRequest.m_sClose = "NV_SESSION_CLOSE_NONE";
	SpecialRequest.m_sUser = UserNameHere;
	if(UserPassword != NULL)
		SpecialRequest.m_sPassword = UserPassword;
	if(NewPassword != NULL)
		SpecialRequest.m_sNewPassword = NewPassword;
	if(NewPasswordConfirm != NULL)
		SpecialRequest.m_sNewPasswordConfirm = NewPasswordConfirm;
	SpecialRequest.bWithNewPassword = bWithNewPassword;
	SpecialRequest.bWithNewPasswordConfirm = bWithNewPasswordConfirm;


	CleanString(ServerAdress);
	CleanString(SpecialRequest.m_sAppli);
	CleanString(SpecialRequest.m_sSessionId);
	CleanString(SpecialRequest.m_sUser);
	CleanString(SpecialRequest.m_sPassword);
	SpecialRequest.m_sHash = SpecialRequest.m_sAppli + SpecialRequest.m_sUser + SpecialRequest.m_sPassword;
	SpecialRequest.m_sHash = MD5_string(SpecialRequest.m_sHash);
	MakeUpper(SpecialRequest.m_sHash);

	if(!ServerAdress.length())
		ServerAdress = "127.0.0.1";

	bool bVirtualModeLoadBalancing = false;
	int iVirtualNumAddresses = 0;
	if((ServerAdress[0] == '(') && (ServerAdress[ServerAdress.length()-1] == ')'))
	{
		// This is a virtual host address
		ServerAdress = ServerAdress.substr(1); 
		ServerAdress = ServerAdress.substr(0, ServerAdress.length()-1); 
		iVirtualNumAddresses = GetMain()->VirtualHosts.GetNumAddresses(ServerAdress.c_str());
		if(iVirtualNumAddresses != 0)
		{
			if(GetMain()->VirtualHosts.GetMode(ServerAdress.c_str()) == "LOAD_BALANCING")
				bVirtualModeLoadBalancing = true;
		}
	}
	if(!iVirtualNumAddresses)
		SpecialRequest.m_sServer = ServerAdress;

	SpecialRequest.m_iAutoClose = NV_REQUEST_CLOSE_SESSION;
	SpecialRequest.m_iSessionTimeOut = 60;
	SpecialRequest.m_iServerConnectionTimeOut = 10;
	SpecialRequest.b_Hidden = true;
	SpecialRequest.bControlPassword = ControlPassword;

	int RetVal = 1;

	if(iVirtualNumAddresses)
	{
		// We must get the real address and test the server
		NvCommand TestCommand;
		TestCommand.m_pSession = this;
		TestCommand.m_iCommandSource = NV_SOURCE_CLIENT;
		TestCommand.m_sService = "SYSTEM";
		TestCommand.m_sClass = "MISC";
		TestCommand.m_sCommand = "NOP";
		bool bConnected = false;
		if(!bVirtualModeLoadBalancing)
		{
			// In failover mode, we first try to connect the active server
			SpecialRequest.m_sServer = GetMain()->VirtualHosts.GetActiveAddress(ServerAdress.c_str());
			if(!SpecialRequest.m_sServer.length())
				SpecialRequest.m_sServer = "127.0.0.1";
			if(SpecialRequest.RequestCommand(&TestCommand))
				bConnected = true;
			else
			{
				// If the error is not a network error
				// the peer server is running so we must return the error
				if((TestCommand.Error.m_sErrorService != "SYSTEM") ||
					(TestCommand.Error.m_sErrorClass != "REQUEST") ||
					(TestCommand.Error.m_iErrorCode != NV_REQUEST_NETWORK_ERROR))
				{
					RetVal = 0;
					if(TestCommand.Error.m_iErrorCode == NV_SECURITY_BAD_USER)
						RetVal = -1;
					else if(TestCommand.Error.m_iErrorCode == NV_SECURITY_CANNOT_GET_CONTEXT)
						RetVal = -2;
					else if(TestCommand.Error.m_iErrorCode == NV_SECURITY_BAD_PASSWORD)
						RetVal = -3;
					else if(TestCommand.Error.m_iErrorCode == NV_SECURITY_BAD_APPLICATION)
						RetVal = -4;
					else if(TestCommand.Error.m_iErrorCode == NV_SECURITY_CHANGE_PASSWORD)
						RetVal = -5;
					else if(TestCommand.Error.m_iErrorCode == NV_SECURITY_BAD_PASSWORD_SYNTAX)
						RetVal = -6;
					SpecialRequest.Clear();
					delete SpecialSession; 
					return RetVal;
				}
			}
		}
		while(!bConnected)
		{
			// We must try a connection to next address but no more than the number of addresses defined
			if(!iVirtualNumAddresses)
				break;
			iVirtualNumAddresses--;
			SpecialRequest.m_sServer = GetMain()->VirtualHosts.GetNextAddress(ServerAdress.c_str());
			if(!SpecialRequest.m_sServer.length())
				SpecialRequest.m_sServer = "127.0.0.1";
			if(SpecialRequest.RequestCommand(&TestCommand))
				bConnected = true;
			else
			{
				// If the error is not a network error
				// the peer server is running so we must return the error
				if((TestCommand.Error.m_sErrorService != "SYSTEM") ||
					(TestCommand.Error.m_sErrorClass != "REQUEST") ||
					(TestCommand.Error.m_iErrorCode != NV_REQUEST_NETWORK_ERROR))
				{
					RetVal = 0;
					if(TestCommand.Error.m_iErrorCode == NV_SECURITY_BAD_USER)
						RetVal = -1;
					else if(TestCommand.Error.m_iErrorCode == NV_SECURITY_CANNOT_GET_CONTEXT)
						RetVal = -2;
					else if(TestCommand.Error.m_iErrorCode == NV_SECURITY_BAD_PASSWORD)
						RetVal = -3;
					else if(TestCommand.Error.m_iErrorCode == NV_SECURITY_BAD_APPLICATION)
						RetVal = -4;
					else if(TestCommand.Error.m_iErrorCode == NV_SECURITY_CHANGE_PASSWORD)
						RetVal = -5;
					else if(TestCommand.Error.m_iErrorCode == NV_SECURITY_BAD_PASSWORD_SYNTAX)
						RetVal = -6;
					SpecialRequest.Clear();
					delete SpecialSession; 
					return RetVal;
				}
			}
		}
		if(!bConnected)
		{
			SpecialRequest.Clear();
			delete SpecialSession; 
			return 0;
		}
	}


	NvCommand Command;
	SpecialSession->OriginalCommand = &Command;
	Command.m_pSession = SpecialSession;
	Command.m_iCommandSource = NV_SOURCE_CLIENT;
	Command.m_sService = "SYSTEM";
	Command.m_sClass = "SECURITY";
	Command.m_sCommand = "SESSION_PERMISSIONS";
	Command.Parameters.SetValue("APPONLY", "NO");
	SpecialSession->Application = NULL;

	if(!SpecialRequest.RequestCommand(&Command))
	{
		RetVal = 0;
		string ErrorService = Command.Error.m_sErrorService;
		string ErrorClass = Command.Error.m_sErrorClass;
		if((ErrorService == "SYSTEM") && (ErrorClass == "SECURITY"))
		{
			if(Command.Error.m_iErrorCode == NV_SECURITY_BAD_USER)
				RetVal = -1;
			else if(Command.Error.m_iErrorCode == NV_SECURITY_CANNOT_GET_CONTEXT)
				RetVal = -2;
			else if(Command.Error.m_iErrorCode == NV_SECURITY_BAD_PASSWORD)
				RetVal = -3;
			else if(Command.Error.m_iErrorCode == NV_SECURITY_BAD_APPLICATION)
				RetVal = -4;
			else if(Command.Error.m_iErrorCode == NV_SECURITY_CHANGE_PASSWORD)
				RetVal = -5;
			else if(Command.Error.m_iErrorCode == NV_SECURITY_BAD_PASSWORD_SYNTAX)
				RetVal = -6;
		}
	}
	if(RetVal != 1)
	{
		SpecialRequest.Clear();
		delete SpecialSession; 
		return RetVal;
	}


	// Now we are connected to the distant application
	// so we must issue a command for retreiving the list of permissions
	Command.m_pSession = SpecialSession;
	Command.m_iCommandSource = NV_SOURCE_CLIENT;
	Command.m_sService = "SYSTEM";
	Command.m_sClass = "OBJECT";
	Command.m_sCommand = "GET";
	Command.Parameters.RemoveAll();
	Command.Parameters.SetValue("NAME", "PERMISSION_LIST");
	NvContainer *Container = SpecialSession->GetContainer("");
	SpecialSession->InContainer = Container;
	if(!SpecialRequest.RequestCommand(&Command))
	{
		SpecialRequest.Clear();
		SpecialSession->FreeContainer(Container, "");
		delete SpecialSession; 
		return 0;
	}


	// We have now the list of permissions from peer server
	// so we just have to fill our own list
	NvObjStringList *PermissionList = (NvObjStringList*)(Container->GetObject("PERMISSION_LIST"));
	if(PermissionList != NULL)
	{
		int NumInList = PermissionList->GetSize();
		int i;
		for(i=0;i<NumInList;i++)
			UserPermissions->SetValue(PermissionList->GetValue(i), "");
	}


	// get now the user full name
	Command.m_pSession = SpecialSession;
	Command.m_iCommandSource = NV_SOURCE_CLIENT;
	Command.m_sService = "SYSTEM";
	Command.m_sClass = "SESSION";
	Command.m_sCommand = "GET_USER";
	Command.Parameters.RemoveAll();
	Command.Parameters.SetValue("ALL_INFO", "YES");

	if(SpecialRequest.RequestCommand(&Command))
	{
		Command.m_pSession = SpecialSession;
		Command.m_iCommandSource = NV_SOURCE_CLIENT;
		Command.m_sService = "SYSTEM";
		Command.m_sClass = "OBJECT";
		Command.m_sCommand = "GET";
		Command.Parameters.RemoveAll();
		Command.Parameters.SetValue("NAME", "USER_INFO");
		NvContainer *Container = SpecialSession->GetContainer("");
		SpecialSession->InContainer = Container;
		if(SpecialRequest.RequestCommand(&Command))
		{
			NvObjIndStringList *UserFullNameObj = (NvObjIndStringList*)(Container->GetObject("USER_INFO"));
			if(UserFullNameObj != NULL)
				UserFullName = UserFullNameObj->GetValue("FULLNAME");
		}
	}

	SpecialRequest.Clear();

	// Do cleanup now
	SpecialSession->FreeContainer(Container, "");
	delete SpecialSession; 
	
	// The login is OK
	NV_LOCKOBJECT(this);
	Sessions.SetValue(SessionIdHere, UserNameHere.c_str());
	NV_UNLOCKOBJECT(this);

	return 1;
}


// Do the login on by the way of dedicated service
// return 1 if OK, 0 for general error, -1 if bad user, -2 if cannot get the security context, -3 if bad password, -4 if application not enabled, -5 if password must be changed
int NvApplication::ServiceLogin(const char *SessionId, const char *UserName, const char *UserPassword, const char *NewPassword, const char *NewPasswordConfirm, bool bWithNewPassword, bool bWithNewPasswordConfirm, NvObjIndStringList *UserPermissions, string& UserFullName, bool ControlPassword)
{
	string SecurityService = GetServiceSecurity();
	UserFullName = "";

	string UserNameHere = "";
	if(UserName != NULL)
		UserNameHere = UserName;
	string UserPasswordHere = "";
	if(UserPassword != NULL)
		UserPasswordHere = UserPassword;
	string NewPasswordHere = "";
	if(NewPassword != NULL)
		NewPasswordHere = NewPassword;
	if(!bWithNewPassword)
		NewPasswordHere = "";
	string NewPasswordConfirmHere = "";
	if(NewPasswordConfirm != NULL)
		NewPasswordConfirmHere = NewPasswordConfirm;
	if(!bWithNewPasswordConfirm)
		NewPasswordConfirmHere = "";
	if(!ControlPassword)
	{
		UserPasswordHere = "";
		NewPasswordHere = "";
		NewPasswordConfirmHere = "";
		bWithNewPassword = false;
		bWithNewPasswordConfirm = false;
	}

	if(bWithNewPasswordConfirm)
	{
		if(NewPasswordHere != NewPasswordConfirmHere)
		{
			// The password must be identical
			return -3;
		}
	}

	// Create a new temporary session just for setting password
	NvSession *SpecialSession = NVNEW NvSession;
	SpecialSession->Init();
	SpecialSession->m_bAdministrator = true;
	SpecialSession->bHidden = true;
	SpecialSession->Application = this;

	// We must set the InOpenProc flag because otherwise
	// the service will not be able to set security permissions
	SpecialSession->bInOpenProc = true;

	NvCommand Command;
	Command.m_pSession = SpecialSession;
	Command.m_iCommandSource = NV_SOURCE_SERVER;
	Command.m_sService = SecurityService;
	Command.m_sClass = "SECURITY";
	Command.m_sCommand = "LOGIN";
	Command.Parameters.SetValue("USER", UserName);
	Command.Parameters.SetValue("PASSWORD", UserPassword);
	Command.Parameters.SetValue("NEW_PASSWORD", NewPasswordHere);
	if(bWithNewPassword)
		Command.Parameters.SetValue("HAVE_NEW_PASSWORD", "YES");
	else
		Command.Parameters.SetValue("HAVE_NEW_PASSWORD", "NO");
	if(ControlPassword)
		Command.Parameters.SetValue("CONTROL_PASSWORD", "YES");
	else
		Command.Parameters.SetValue("CONTROL_PASSWORD", "NO");
	Command.Parameters.SetValue("APPLICATION", GetName());

	int RetVal = 1;
	if(SpecialSession->Command(&Command))
	{
		// The command was successful
		// We must get the eventual user full name and the user permissions
		NvContainer* Container = SpecialSession->GetContainer("");
		if(Container != NULL)
		{
			NvObjString* ObjUserFullName = (NvObjString*)(Container->GetObject("USER_FULL_NAME"));
			if(ObjUserFullName != NULL)
				UserFullName = ObjUserFullName->GetValue();
		}

		// Now get permissions
		UserPermissions->CopyFrom(&(SpecialSession->m_oPermissions));
	}
	else
	{
		// Command failed
		RetVal = 0;
		string ErrorClass = Command.Error.m_sErrorClass;
		if(ErrorClass == "SECURITY")
		{
			if(Command.Error.m_iErrorCode == NV_SECURITY_BAD_USER)
				RetVal = -1;
			else if(Command.Error.m_iErrorCode == NV_SECURITY_CANNOT_GET_CONTEXT)
				RetVal = -2;
			else if(Command.Error.m_iErrorCode == NV_SECURITY_BAD_PASSWORD)
				RetVal = -3;
			else if(Command.Error.m_iErrorCode == NV_SECURITY_BAD_APPLICATION)
				RetVal = -4;
			else if(Command.Error.m_iErrorCode == NV_SECURITY_CHANGE_PASSWORD)
				RetVal = -5;
			else if(Command.Error.m_iErrorCode == NV_SECURITY_BAD_PASSWORD_SYNTAX)
				RetVal = -6;
		}
	}

	// Do cleanup now
	delete SpecialSession; 

	return RetVal;
}


bool NvApplication::IsLocalSecurity()
{
	string ActualSecuApp = GetSecuApp();
	MakeUpper(ActualSecuApp);
	if(ActualSecuApp.length())
	{
		if((ActualSecuApp != "(SYSTEM)") && (ActualSecuApp.find('[') != string::npos) && (ActualSecuApp.find(']') != string::npos))
			return false;
	}
	return true;
}

string NvApplication::GetServiceSecurity()
{
	string SecurityService = "";
	string ActualSecuApp = GetSecuApp();
	MakeUpper(ActualSecuApp);
	CleanString(ActualSecuApp);
	if(ActualSecuApp.length())
	{
		if((ActualSecuApp[0] == '(') && (ActualSecuApp[ActualSecuApp.length()-1] == ')'))
		{
			SecurityService = ActualSecuApp.substr(1);
			SecurityService = SecurityService.substr(0, SecurityService.length()-1);
			CleanString(SecurityService);
			if(SecurityService == "SYSTEM")
				SecurityService = "";
		}
	}
	return SecurityService;
}

bool NvApplication::Logout(const char *SessionId)
{
	//PM TODO

	string SessionIdHere = SessionId;
	CleanString(SessionIdHere);
	MakeUpper(SessionIdHere);
	
	NV_LOCKOBJECT(this);
	Sessions.Remove(SessionIdHere);
	NV_UNLOCKOBJECT(this);

	SemaList.UnlockAll(SessionId);
	LockList.UnlockAll(SessionId);

	return true;
}


NvSecurity* NvApplication::GetSecurity(string& AppName)
{
	AppName = "";
	NvSecurity *CurrentSecurity = &Security;
	NvApplication *Application = NULL;
	string ActualSecuApp = GetSecuApp();
	if(ActualSecuApp.length())
	{
		if(ActualSecuApp == "(SYSTEM)")
			CurrentSecurity = &(GetMain()->Security);
		else
		{
			// Other application security
			if(!GetMain()->Applications.IsApplicationEnable(ActualSecuApp.c_str()))
				return NULL;	// appli not enabled

			Application = GetMain()->Applications.GetApplication(ActualSecuApp.c_str());

			// The application doesn't exist
			if(Application == NULL)
				return NULL;
			CurrentSecurity = &(Application->Security);
			AppName = Application->GetName();
		}
	}

	return CurrentSecurity;
}



bool NvApplication::Package(const char *DescriptionFile, const char *PackageFile, string& Info)
{

	NV_LOCKLOCK(&ProtectInUse);
	if(!CreateApplicationDirectories())
	{
		NV_UNLOCKLOCK(&ProtectInUse);
		return false;
	}
	NV_UNLOCKLOCK(&ProtectInUse);

	string AppDirName = GetMain()->GetNirvaDir();
	AppDirName += "Applications";
	AddDirSuffix(AppDirName);
	AppDirName += m_sName;
	AddDirSuffix(AppDirName);

	string DescriptionFileInParam = "";
	if(DescriptionFile != NULL)
		DescriptionFileInParam = DescriptionFile;
	CleanString(DescriptionFileInParam);

	string DescriptionFileHere = AppDirName;
	DescriptionFileHere += "Files";
	AddDirSuffix(DescriptionFileHere);
	if(DescriptionFileInParam.length())
		DescriptionFileHere += DescriptionFileInParam;
	else
		DescriptionFileHere += "package.lst";

	NvPackage Package;
	bool RetVal = Package.Check(DescriptionFileHere.c_str(), AppDirName.c_str());
	if(RetVal)
		RetVal = Package.Create(DescriptionFileHere.c_str(), AppDirName.c_str(), PackageFile);
	if(!RetVal)
		Info = Package.Info;

	return RetVal;
}

bool NvApplication::Install(const char *PackageFile, string& Info)
{
	NV_LOCKLOCK(&ProtectInUse);
	if(!CreateApplicationDirectories())
	{
		NV_UNLOCKLOCK(&ProtectInUse);
		NV_UNLOCKOBJECT(this);
		return false;
	}
	NV_UNLOCKLOCK(&ProtectInUse);

	string AppDirName = GetMain()->GetNirvaDir();
	AppDirName += "Applications";
	AddDirSuffix(AppDirName);
	AppDirName += m_sName;
	AddDirSuffix(AppDirName);

	NvPackage Package;
	bool RetVal = Package.Install(PackageFile, AppDirName.c_str(), GetMain()->GetDebug());
	if(!RetVal)
		Info = Package.Info;

	return RetVal;
}


/////////////////////////////////////////////////////////////////////////////
// NvRedirection class
NvRedirection::NvRedirection()
{
	m_sAddress = "";
	m_bEnable = true;
}

NvRedirection::~NvRedirection()
{
}




/////////////////////////////////////////////////////////////////////////////
// NvRedirectionList

NvRedirectionList::NvRedirectionList()
{
}

NvRedirectionList::~NvRedirectionList()
{
	NV_LOCKOBJECT(this);
	Clear();
	NV_UNLOCKOBJECT(this);
}



void NvRedirectionList::Clear()
{
	map<string,NvRedirection*>::iterator pos;
	for(pos = m_mRedirections.begin(); pos != m_mRedirections.end(); ++pos)
	{
		NvRedirection *Redirection = pos->second;
		if(Redirection != NULL)
			delete Redirection;
	}
	m_mRedirections.clear();
}

NvRedirection* NvRedirectionList::InternalGetRedirection(const char *Address)
{
	if(Address == NULL)
		return NULL;

	string AddressHere = Address;
	CleanString(AddressHere);

	if(!AddressHere.length())
		return NULL;

	if(m_mRedirections.find(AddressHere) == m_mRedirections.end())
		return NULL;
	
	NvRedirection *FoundRedirection = m_mRedirections[AddressHere];

	return FoundRedirection;
}

// Load
bool NvRedirectionList::Load()
{
	NV_LOCKOBJECT(this);
	Clear();

	// load the redirection table
	// The redirection table is stored in a table object named "redirections" in the
	// application registry key system.parameters
	NvContainer Container;
	if(Application->AppRegistry.GetKey("system.parameters", &Container, NV_CONTAINER_COPY_NO_APPEND, "redirections", false, true))
	{
		NvObj *Object = Container.GetObject("redirections");
		if((Object != NULL) && (Object->GetType() == NV_OBJTYPE_TABLE))
		{
			NvObjTable *RedirectionTable = (NvObjTable*)Object;
			int index_Address = RedirectionTable->GetColumnIndex("Address");
			int index_Enable = RedirectionTable->GetColumnIndex("Enable");
			int row_index;
			for(row_index=0; row_index<RedirectionTable->GetNumRow(); row_index++)
			{
				string Address = RedirectionTable->GetCellLine(row_index, index_Address);
				CleanString(Address);
				if(!Address.length())
					continue;
				
				if(InternalGetRedirection(Address.c_str()) != NULL)	// Check doubles
					continue;
				string RedirectionEnable = RedirectionTable->GetCellLine(row_index, index_Enable);
				bool bRedirectionEnable = true;
				if(RedirectionEnable.length() && (atoi(RedirectionEnable.c_str()) == 0))
					bRedirectionEnable = false;
				
				// Create the redirection in redirection list
				if(AddRedirection(Address.c_str(), false))
					EnableRedirection(Address.c_str(), bRedirectionEnable, false);
			}
		}
	}
	
	// Saves the list in case of changes
	Save();

	NV_UNLOCKOBJECT(this);
	
	return true;
}

// Save
bool NvRedirectionList::Save()
{
	// First create the registry key "system.parameters" if it doesn't exist
	Application->AppRegistry.SetKey("system.parameters", NULL);

	NvContainer Container;
	if(!Application->AppRegistry.GetKey("system.parameters", &Container, NV_CONTAINER_COPY_NO_APPEND, NULL, false, true))
		return false; // cannot get the key. This should not occur here

	// Creates the redirections object if it doesn't exist
	Container.CreateObject("redirections", NV_OBJTYPE_TABLE);
	NvObj *Object = Container.GetObject("redirections");
	if(Object == NULL)
		return false; // This should not occur here

	if(Object->GetType() != NV_OBJTYPE_TABLE)
	{
		Container.RemoveObject("redirections");
		Container.CreateObject("redirections", NV_OBJTYPE_TABLE);
		Object = Container.GetObject("redirections");
		if(Object == NULL)
			return false; // This should not occur here
	}

	NvObjTable *RedirectionTable = (NvObjTable*)Object;
	RedirectionTable->Clear();
	RedirectionTable->InsertColumn("Address");
	RedirectionTable->InsertColumn("Enable");

	int index_Address = RedirectionTable->GetColumnIndex("Address");
	int index_Enable = RedirectionTable->GetColumnIndex("Enable");

	int j;
	map<string,NvRedirection*>::iterator pos;
	for(j=0, pos = m_mRedirections.begin(); pos != m_mRedirections.end(); ++pos)
	{
		NvRedirection *Redirection = pos->second;
		if(Redirection == NULL)
			continue;

		RedirectionTable->InsertRow();
		RedirectionTable->InsertCellLine(j, index_Address, Redirection->m_sAddress.c_str());
		if(Redirection->m_bEnable)
			RedirectionTable->InsertCellLine(j, index_Enable, "1");
		else
			RedirectionTable->InsertCellLine(j, index_Enable, "0");
		j++;
	}
	
	if(!Application->AppRegistry.SetKey("system.parameters", &Container))
		return false;

	return true;
}

// Get list for http
bool NvRedirectionList::GetRedirections(std::vector<std::string> &List)
{
	NV_LOCKOBJECT(this);
	map<string,NvRedirection*>::iterator pos;
	for(pos = m_mRedirections.begin(); pos != m_mRedirections.end(); ++pos)
	{
		NvRedirection *Redirection = pos->second;
		if(Redirection == NULL)
			continue;
		if(!Redirection->m_bEnable)
			continue;
		List.push_back(Redirection->m_sAddress);
	}
	NV_UNLOCKOBJECT(this);
	return true;
}

// Get redirection info
bool NvRedirectionList::GetRedirectionInfo(const char *Address, NvObjTable *RedirectionInfo)
{
	RedirectionInfo->Clear();

	bool AllRedirections = true;

	NV_LOCKOBJECT(this);

	NvRedirection *FoundRedirection = NULL;
	if((Address != NULL) && strlen(Address))
	{
		// There is a request on a specific set
		FoundRedirection = InternalGetRedirection(Address);
		if(FoundRedirection == NULL)
		{
			NV_UNLOCKOBJECT(this);
			return false;
		}
		AllRedirections = false;
	}


	RedirectionInfo->InsertColumn("ADDRESS");
	RedirectionInfo->InsertColumn("ENABLE");

	if(AllRedirections)
	{
		int j;
		map<string,NvRedirection*>::iterator pos;
		for(pos = m_mRedirections.begin(), j=0; pos != m_mRedirections.end(); ++pos)
		{
			NvRedirection *Redirection = pos->second;
			if(Redirection == NULL)
				continue;
			RedirectionInfo->InsertRow();
			RedirectionInfo->InsertCellLine(j, 0, Redirection->m_sAddress);
			if(Redirection->m_bEnable)
				RedirectionInfo->InsertCellLine(j, 1, "YES");
			else
				RedirectionInfo->InsertCellLine(j, 1, "NO");
			j++;
		}
	}
	else
	{
		RedirectionInfo->InsertRow();
		RedirectionInfo->InsertCellLine(0, 0, FoundRedirection->m_sAddress);
		if(FoundRedirection->m_bEnable)
			RedirectionInfo->InsertCellLine(0, 1, "YES");
		else
			RedirectionInfo->InsertCellLine(0, 1, "NO");
	}
	NV_UNLOCKOBJECT(this);

	return true;
}

// Add a redirection
bool NvRedirectionList::AddRedirection(const char *Address, bool bLockIt)
{
	string AddressHere = "";
	if(Address != NULL)
		AddressHere = Address;
	CleanString(AddressHere);
	if(!AddressHere.length())
		return false;

	if(bLockIt)
		NV_LOCKOBJECT(this);

	if(InternalGetRedirection(Address) != NULL)
	{
		if(bLockIt)
			NV_UNLOCKOBJECT(this);
		return false;
	}

	NvRedirection *Redirection = new NvRedirection;
	Redirection->m_sAddress = AddressHere;
	m_mRedirections[AddressHere] = Redirection;
	if(bLockIt)
	{
		Save(); // We save if we have to lock because this comes from a command and not from Load
		NV_UNLOCKOBJECT(this);
	}
	return true;
}

// Remove a redirection
bool NvRedirectionList::RemoveRedirection(const char *Address)
{
	NV_LOCKOBJECT(this);
	NvRedirection *Redirection = InternalGetRedirection(Address);
	if(Redirection != NULL)
	{
		delete Redirection;
		m_mRedirections.erase(Address);
		Save();
	}
	NV_UNLOCKOBJECT(this);
	return true;
}

// Enable a redirection
bool NvRedirectionList::EnableRedirection(const char *Address, bool Enable, bool bLockIt)
{
	if(bLockIt)
		NV_LOCKOBJECT(this);
	NvRedirection *Redirection = InternalGetRedirection(Address);
	if(Redirection != NULL)
		Redirection->m_bEnable = Enable;
	if(bLockIt)
	{
		Save(); // We save if we have to lock because this comes from a command and not from Load
		NV_UNLOCKOBJECT(this);
	}
	return true;
}

